from collections.abc import Mapping
from typing import NamedTuple

import dagster._check as check
from dagster._annotations import PublicAttr
from dagster._core.definitions import ExecutorDefinition, IJob
from dagster._core.instance import DagsterInstance


class InitExecutorContext(
    NamedTuple(
        "InitExecutorContext",
        [
            ("job", PublicAttr[IJob]),
            ("executor_def", PublicAttr[ExecutorDefinition]),
            ("executor_config", PublicAttr[Mapping[str, object]]),
            ("instance", PublicAttr[DagsterInstance]),
        ],
    )
):
    """Executor-specific initialization context.

    Args:
        job (IJob): The job to be executed.
        executor_def (ExecutorDefinition): The definition of the executor currently being
            constructed.
        executor_config (dict): The parsed config passed to the executor.
        instance (DagsterInstance): The current instance.
    """

    def __new__(
        cls,
        job: IJob,
        executor_def: ExecutorDefinition,
        executor_config: Mapping[str, object],
        instance: DagsterInstance,
    ):
        return super().__new__(
            cls,
            job=check.inst_param(job, "job", IJob),
            executor_def=check.inst_param(executor_def, "executor_def", ExecutorDefinition),
            executor_config=check.mapping_param(executor_config, "executor_config", key_type=str),
            instance=check.inst_param(instance, "instance", DagsterInstance),
        )



import importlib
import importlib.metadata
import inspect
import sys
from collections.abc import Iterable, Sequence
from types import ModuleType

from dagster_shared.serdes.objects import LibraryObjectKey

from dagster._core.errors import DagsterError
from dagster.components.utils import format_error_message

LIBRARY_OBJECT_ATTR = "__dg_library_object__"
DG_LIBRARY_ENTRY_POINT_GROUP = "dagster_dg.library"


class ComponentsEntryPointLoadError(DagsterError):
    pass


def get_entry_points_from_python_environment(group: str) -> Sequence[importlib.metadata.EntryPoint]:
    if sys.version_info >= (3, 10):
        return importlib.metadata.entry_points(group=group)
    else:
        return importlib.metadata.entry_points().get(group, [])


def discover_entry_point_library_objects() -> dict[LibraryObjectKey, object]:
    """Discover library objects registered in the Python environment via the
    `dg_library` entry point group.

    `dagster-components` itself registers multiple component entry points. We call these
    "builtin" component libraries. The `dagster_components` entry point resolves to published
    component types and is loaded by default. Other entry points resolve to various sets of test
    component types. This method will only ever load one builtin component library.
    """
    objects: dict[LibraryObjectKey, object] = {}
    entry_points = get_entry_points_from_python_environment(DG_LIBRARY_ENTRY_POINT_GROUP)

    for entry_point in entry_points:
        try:
            root_module = entry_point.load()
        except Exception as e:
            raise ComponentsEntryPointLoadError(
                format_error_message(f"""
                    Error loading entry point `{entry_point.name}` in group `{DG_LIBRARY_ENTRY_POINT_GROUP}`.
                    Please fix the error or uninstall the package that defines this entry point.
                """)
            ) from e

        if not isinstance(root_module, ModuleType):
            raise DagsterError(
                f"Invalid entry point {entry_point.name} in group {DG_LIBRARY_ENTRY_POINT_GROUP}. "
                f"Value expected to be a module, got {root_module}."
            )
        for name, obj in get_library_objects_in_module(root_module):
            key = LibraryObjectKey(name=name, namespace=entry_point.value)
            objects[key] = obj
    return objects


def discover_library_objects(modules: Sequence[str]) -> dict[LibraryObjectKey, object]:
    objects: dict[LibraryObjectKey, object] = {}
    for extra_module in modules:
        for name, obj in get_library_objects_in_module(importlib.import_module(extra_module)):
            key = LibraryObjectKey(name=name, namespace=extra_module)
            objects[key] = obj
    return objects


def get_library_objects_in_module(
    module: ModuleType,
) -> Iterable[tuple[str, object]]:
    for attr in dir(module):
        value = getattr(module, attr)
        if hasattr(value, LIBRARY_OBJECT_ATTR) and not inspect.isabstract(value):
            yield attr, value


def load_library_object(key: LibraryObjectKey) -> object:
    module_name, attr = key.namespace, key.name
    try:
        module = importlib.import_module(module_name)
        if not hasattr(module, attr):
            raise DagsterError(f"Module `{module_name}` has no attribute `{attr}`.")
        return getattr(module, attr)
    except ModuleNotFoundError as e:
        raise DagsterError(f"Module `{module_name}` not found.") from e
    except ImportError as e:
        raise DagsterError(f"Error loading module `{module_name}`.") from e



import contextlib
import contextvars
import dataclasses
import importlib
from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, Union

from dagster_shared.yaml_utils.source_position import SourcePositionTree

from dagster._core.definitions.definitions_class import Definitions
from dagster._core.errors import DagsterError
from dagster._utils import pushd
from dagster.components.resolved.context import ResolutionContext
from dagster.components.utils import get_path_from_module


@dataclass
class ComponentLoadContext:
    """Context for loading a single component."""

    path: Path
    defs_module_path: Path
    defs_module_name: str
    resolution_context: ResolutionContext

    @staticmethod
    def current() -> "ComponentLoadContext":
        context = active_component_load_context.get()
        if context is None:
            raise DagsterError(
                "No active component load context, `ComponentLoadContext.current()` must be called inside of a component's `build_defs` method"
            )
        return context

    @staticmethod
    def for_module(defs_module: ModuleType) -> "ComponentLoadContext":
        path = get_path_from_module(defs_module)
        return ComponentLoadContext(
            path=path,
            defs_module_path=path,
            defs_module_name=defs_module.__name__,
            resolution_context=ResolutionContext.default(),
        )

    @staticmethod
    def for_test() -> "ComponentLoadContext":
        return ComponentLoadContext(
            path=Path.cwd(),
            defs_module_path=Path.cwd(),
            defs_module_name="test",
            resolution_context=ResolutionContext.default(),
        )

    def _with_resolution_context(
        self, resolution_context: ResolutionContext
    ) -> "ComponentLoadContext":
        return dataclasses.replace(self, resolution_context=resolution_context)

    def with_rendering_scope(self, rendering_scope: Mapping[str, Any]) -> "ComponentLoadContext":
        return self._with_resolution_context(self.resolution_context.with_scope(**rendering_scope))

    def with_source_position_tree(
        self, source_position_tree: SourcePositionTree
    ) -> "ComponentLoadContext":
        return self._with_resolution_context(
            self.resolution_context.with_source_position_tree(source_position_tree)
        )

    def for_path(self, path: Path) -> "ComponentLoadContext":
        return dataclasses.replace(self, path=path)

    def defs_relative_module_name(self, path: Path) -> str:
        """Returns the name of the python module at the given path, relative to the project root."""
        container_path = self.path.parent if self.path.is_file() else self.path
        with pushd(str(container_path)):
            relative_path = path.resolve().relative_to(self.defs_module_path.resolve())
            if path.name == "__init__.py":
                # e.g. "a_project/defs/something/__init__.py" -> "a_project.defs.something"
                relative_parts = relative_path.parts[:-1]
            elif path.is_file():
                # e.g. "a_project/defs/something/file.py" -> "a_project.defs.something.file"
                relative_parts = [*relative_path.parts[:-1], relative_path.stem]
            else:
                # e.g. "a_project/defs/something/" -> "a_project.defs.something"
                relative_parts = relative_path.parts
            return ".".join([self.defs_module_name, *relative_parts])

    def normalize_component_type_str(self, type_str: str) -> str:
        return (
            f"{self.defs_relative_module_name(self.path)}{type_str}"
            if type_str.startswith(".")
            else type_str
        )

    def load_defs(self, module: ModuleType) -> Definitions:
        """Builds the set of Dagster definitions for a component module.

        This is useful for resolving dependencies on other components.
        """
        # FIXME: This should go through the component loader system
        # to allow for this value to be cached and more selectively
        # loaded. This is just a temporary hack to keep tests passing.
        from dagster.components.core.load_defs import load_defs

        return load_defs(module)

    def load_defs_relative_python_module(self, path: Path) -> ModuleType:
        """Load a python module relative to the defs's context path. This is useful for loading code
        the resides within the defs directory.

        Example:
            .. code-block:: python

                def build_defs(self, context: ComponentLoadContext) -> Definitions:
                    return load_definitions_from_module(
                        context.load_defs_relative_python_module(
                            Path(self.definitions_path) if self.definitions_path else Path("definitions.py")
                        )
                    )

        In a typical setup you end up with module names such as "a_project.defs.my_component.my_python_file".

        This handles "__init__.py" files by ending the module name at the parent directory
        (e.g "a_project.defs.my_component") if the file resides at "a_project/defs/my_component/__init__.py".

        This calls importlib.import_module with that module name, going through the python module import system.

        It is as if one typed "import a_project.defs.my_component.my_python_file" in the python interpreter.
        """
        return importlib.import_module(self.defs_relative_module_name(path))


active_component_load_context: contextvars.ContextVar[Union[ComponentLoadContext, None]] = (
    contextvars.ContextVar("active_component_load_context", default=None)
)


@contextlib.contextmanager
def use_component_load_context(component_load_context: ComponentLoadContext):
    token = active_component_load_context.set(component_load_context)
    try:
        yield
    finally:
        active_component_load_context.reset(token)



import asyncio
from abc import abstractmethod
from typing import TYPE_CHECKING, AbstractSet, Any, Optional, Sequence  # noqa: UP035

from dagster_shared.serdes import whitelist_for_serdes

import dagster._check as check
from dagster._core.definitions.asset_key import AssetCheckKey, AssetKey
from dagster._core.definitions.base_asset_graph import BaseAssetGraph, BaseAssetNode
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
    BuiltinAutomationCondition,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._core.definitions.declarative_automation.operators.dep_operators import (
    EntityMatchesCondition,
)
from dagster._record import copy, record
from dagster._utils.security import non_secure_md5_hash_str

if TYPE_CHECKING:
    from dagster._core.definitions.asset_selection import AssetSelection


@record
class ChecksAutomationCondition(BuiltinAutomationCondition[AssetKey]):
    operand: AutomationCondition[AssetCheckKey]

    blocking_only: bool = False
    # Should be AssetSelection, but this causes circular reference issues
    allow_selection: Optional[Any] = None
    ignore_selection: Optional[Any] = None

    @property
    @abstractmethod
    def base_name(self) -> str: ...

    @property
    def name(self) -> str:
        name = self.base_name
        props = []
        if self.blocking_only:
            props.append("blocking_only=True")
        if self.allow_selection is not None:
            props.append(f"allow_selection={self.allow_selection}")
        if self.ignore_selection is not None:
            props.append(f"ignore_selection={self.ignore_selection}")

        if props:
            name += f"({','.join(props)})"
        return name

    @property
    def requires_cursor(self) -> bool:
        return False

    def get_node_unique_id(self, *, parent_unique_id: Optional[str], index: Optional[int]) -> str:
        """Ignore allow_selection / ignore_selection for the cursor hash."""
        parts = [str(parent_unique_id), str(index), self.base_name]
        return non_secure_md5_hash_str("".join(parts).encode())

    def get_backcompat_node_unique_ids(
        self, *, parent_unique_id: Optional[str] = None, index: Optional[int] = None
    ) -> Sequence[str]:
        # backcompat for previous cursors where the allow/ignore selection influenced the hash
        return [super().get_node_unique_id(parent_unique_id=parent_unique_id, index=index)]

    def allow(self, selection: "AssetSelection") -> "ChecksAutomationCondition":
        """Returns a copy of this condition that will only consider dependencies within the provided
        AssetSelection.
        """
        from dagster._core.definitions.asset_selection import AssetSelection

        check.inst_param(selection, "selection", AssetSelection)
        allow_selection = (
            selection if self.allow_selection is None else selection | self.allow_selection
        )
        return copy(self, allow_selection=allow_selection)

    def ignore(self, selection: "AssetSelection") -> "ChecksAutomationCondition":
        """Returns a copy of this condition that will ignore dependencies within the provided
        AssetSelection.
        """
        from dagster._core.definitions.asset_selection import AssetSelection

        check.inst_param(selection, "selection", AssetSelection)
        ignore_selection = (
            selection if self.ignore_selection is None else selection | self.ignore_selection
        )
        return copy(self, ignore_selection=ignore_selection)

    def _get_check_keys(
        self, key: AssetKey, asset_graph: BaseAssetGraph[BaseAssetNode]
    ) -> AbstractSet[AssetCheckKey]:
        check_keys = asset_graph.get(key).check_keys
        if self.blocking_only:
            check_keys = {ck for ck in check_keys if asset_graph.get(ck).blocking}
        if self.allow_selection is not None:
            check_keys &= self.allow_selection.resolve_checks(asset_graph)
        if self.ignore_selection is not None:
            check_keys -= self.ignore_selection.resolve_checks(asset_graph)

        return check_keys


@whitelist_for_serdes
@record
class AnyChecksCondition(ChecksAutomationCondition):
    @property
    def base_name(self) -> str:
        return "ANY_CHECKS_MATCH"

    async def evaluate(self, context: AutomationContext[AssetKey]) -> AutomationResult[AssetKey]:  # pyright: ignore[reportIncompatibleMethodOverride]
        true_subset = context.get_empty_subset()

        coroutines = [
            context.for_child_condition(
                child_condition=EntityMatchesCondition(key=check_key, operand=self.operand),
                child_indices=[
                    None,
                    i,
                ],  # Prefer a non-indexed ID in case asset keys move around, but fall back to the indexed one for back-compat
                candidate_subset=context.candidate_subset,
            ).evaluate_async()
            for i, check_key in enumerate(
                sorted(self._get_check_keys(context.key, context.asset_graph))
            )
        ]

        check_results = await asyncio.gather(*coroutines)
        for check_result in check_results:
            true_subset = true_subset.compute_union(check_result.true_subset)

        true_subset = context.candidate_subset.compute_intersection(true_subset)
        return AutomationResult(context, true_subset=true_subset, child_results=check_results)


@whitelist_for_serdes
@record
class AllChecksCondition(ChecksAutomationCondition):
    @property
    def base_name(self) -> str:
        return "ALL_CHECKS_MATCH"

    async def evaluate(self, context: AutomationContext[AssetKey]) -> AutomationResult[AssetKey]:  # pyright: ignore[reportIncompatibleMethodOverride]
        check_results = []
        true_subset = context.candidate_subset

        for i, check_key in enumerate(
            sorted(self._get_check_keys(context.key, context.asset_graph))
        ):
            check_result = await context.for_child_condition(
                child_condition=EntityMatchesCondition(key=check_key, operand=self.operand),
                child_indices=[
                    None,
                    i,
                ],  # Prefer a non-indexed ID in case asset keys move around, but fall back to the indexed one for back-compat
                candidate_subset=context.candidate_subset,
            ).evaluate_async()
            check_results.append(check_result)
            true_subset = true_subset.compute_intersection(check_result.true_subset)

        return AutomationResult(context, true_subset=true_subset, child_results=check_results)



import asyncio
import logging
import sys
from abc import ABC, abstractmethod
from collections.abc import Generator, Iterable, Iterator, Mapping, Sequence
from contextlib import contextmanager
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Generic,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
    cast,
)

import dagster._check as check
from dagster._core.definitions import ExecutorDefinition, JobDefinition
from dagster._core.definitions.executor_definition import check_cross_process_constraints
from dagster._core.definitions.job_base import IJob
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.definitions.resource_definition import ScopedResourcesBuilder
from dagster._core.errors import DagsterError, DagsterUserCodeExecutionError
from dagster._core.events import DagsterEvent, RunFailureReason
from dagster._core.execution.context.logger import InitLoggerContext
from dagster._core.execution.context.system import (
    ExecutionData,
    IPlanContext,
    PlanData,
    PlanExecutionContext,
    PlanOrchestrationContext,
)
from dagster._core.execution.memoization import validate_reexecution_memoization
from dagster._core.execution.plan.plan import ExecutionPlan
from dagster._core.execution.resources_init import (
    get_required_resource_keys_to_init,
    resource_initialization_manager,
)
from dagster._core.execution.retries import RetryMode
from dagster._core.executor.init import InitExecutorContext
from dagster._core.instance import DagsterInstance
from dagster._core.log_manager import DagsterLogManager
from dagster._core.storage.dagster_run import DagsterRun
from dagster._core.system_config.objects import ResolvedRunConfig
from dagster._loggers import default_loggers, default_system_loggers
from dagster._utils import EventGenerationManager
from dagster._utils.error import serializable_error_info_from_exc_info

if TYPE_CHECKING:
    from dagster._core.execution.plan.outputs import StepOutputHandle
    from dagster._core.executor.base import Executor


def initialize_console_manager(
    dagster_run: Optional[DagsterRun], instance: Optional[DagsterInstance] = None
) -> DagsterLogManager:
    # initialize default colored console logger
    loggers = []
    for logger_def, logger_config in default_system_loggers(instance):
        loggers.append(
            logger_def.logger_fn(
                InitLoggerContext(
                    logger_config, logger_def, run_id=dagster_run.run_id if dagster_run else None
                )
            )
        )
    return DagsterLogManager.create(loggers=loggers, dagster_run=dagster_run, instance=instance)


# This represents all the data that is passed *into* context creation process.
# The remainder of the objects generated (e.g. loggers, resources) are created
# using user-defined code that may fail at runtime and result in the emission
# of a pipeline init failure events. The data in this object are passed all
# over the place during the context creation process so grouping here for
# ease of argument passing etc.
class ContextCreationData(NamedTuple):
    job: IJob
    resolved_run_config: ResolvedRunConfig
    dagster_run: DagsterRun
    executor_def: ExecutorDefinition
    instance: DagsterInstance
    resource_keys_to_init: AbstractSet[str]
    execution_plan: ExecutionPlan

    @property
    def job_def(self) -> JobDefinition:
        return self.job.get_definition()

    @property
    def repository_def(self) -> Optional[RepositoryDefinition]:
        return self.job.get_repository_definition()


def create_context_creation_data(
    job: IJob,
    execution_plan: ExecutionPlan,
    run_config: Mapping[str, object],
    dagster_run: DagsterRun,
    instance: DagsterInstance,
) -> "ContextCreationData":
    job_def = job.get_definition()
    resolved_run_config = ResolvedRunConfig.build(job_def, run_config)

    executor_def = job_def.executor_def

    return ContextCreationData(
        job=job,
        resolved_run_config=resolved_run_config,
        dagster_run=dagster_run,
        executor_def=executor_def,
        instance=instance,
        resource_keys_to_init=get_required_resource_keys_to_init(execution_plan, job_def),
        execution_plan=execution_plan,
    )


def create_plan_data(
    context_creation_data: "ContextCreationData", raise_on_error: bool, retry_mode: RetryMode
) -> PlanData:
    return PlanData(
        job=context_creation_data.job,
        dagster_run=context_creation_data.dagster_run,
        instance=context_creation_data.instance,
        execution_plan=context_creation_data.execution_plan,
        raise_on_error=raise_on_error,
        retry_mode=retry_mode,
    )


def create_execution_data(
    context_creation_data: "ContextCreationData",
    scoped_resources_builder: ScopedResourcesBuilder,
) -> ExecutionData:
    return ExecutionData(
        scoped_resources_builder=scoped_resources_builder,
        resolved_run_config=context_creation_data.resolved_run_config,
        job_def=context_creation_data.job_def,
        repository_def=context_creation_data.repository_def,
    )


TContextType = TypeVar("TContextType", bound=IPlanContext)


class ExecutionContextManager(Generic[TContextType], ABC):
    def __init__(
        self,
        event_generator: Iterator[Union[DagsterEvent, TContextType]],
        raise_on_error: Optional[bool] = False,
    ):
        self._manager = EventGenerationManager[TContextType](
            generator=event_generator, object_cls=self.context_type, require_object=raise_on_error
        )

    @property
    @abstractmethod
    def context_type(self) -> type[TContextType]:
        pass

    def prepare_context(self) -> Iterable[DagsterEvent]:  # ode to Preparable
        return self._manager.generate_setup_events()

    def get_context(self) -> TContextType:
        return self._manager.get_object()

    def shutdown_context(self) -> Iterable[DagsterEvent]:
        return self._manager.generate_teardown_events()

    def get_generator(self) -> Generator[Union[DagsterEvent, IPlanContext], None, None]:
        return self._manager.generator


def execution_context_event_generator(
    job: IJob,
    execution_plan: ExecutionPlan,
    run_config: Mapping[str, object],
    dagster_run: DagsterRun,
    instance: DagsterInstance,
    retry_mode: RetryMode,
    scoped_resources_builder_cm: Optional[
        Callable[..., EventGenerationManager[ScopedResourcesBuilder]]
    ] = None,
    raise_on_error: Optional[bool] = False,
    output_capture: Optional[dict["StepOutputHandle", Any]] = None,
) -> Generator[Union[DagsterEvent, PlanExecutionContext], None, None]:
    scoped_resources_builder_cm = cast(
        Callable[..., EventGenerationManager[ScopedResourcesBuilder]],
        check.opt_callable_param(
            scoped_resources_builder_cm,
            "scoped_resources_builder_cm",
            default=resource_initialization_manager,
        ),
    )

    execution_plan = check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
    job_def = job.get_definition()

    run_config = check.mapping_param(run_config, "run_config", key_type=str)
    dagster_run = check.inst_param(dagster_run, "dagster_run", DagsterRun)
    instance = check.inst_param(instance, "instance", DagsterInstance)

    raise_on_error = check.bool_param(raise_on_error, "raise_on_error")

    context_creation_data = create_context_creation_data(
        job,
        execution_plan,
        run_config,
        dagster_run,
        instance,
    )

    log_manager = create_log_manager(context_creation_data)
    resource_defs = job_def.get_required_resource_defs()
    event_loop = asyncio.new_event_loop()
    try:
        resources_manager = scoped_resources_builder_cm(
            resource_defs=resource_defs,
            resource_configs=context_creation_data.resolved_run_config.resources,
            log_manager=log_manager,
            execution_plan=execution_plan,
            dagster_run=context_creation_data.dagster_run,
            resource_keys_to_init=context_creation_data.resource_keys_to_init,
            instance=instance,
            emit_persistent_events=True,
            event_loop=event_loop,
        )
        yield from resources_manager.generate_setup_events()
        scoped_resources_builder = check.inst(
            resources_manager.get_object(), ScopedResourcesBuilder
        )

        execution_context = PlanExecutionContext(
            plan_data=create_plan_data(context_creation_data, raise_on_error, retry_mode),
            execution_data=create_execution_data(context_creation_data, scoped_resources_builder),
            log_manager=log_manager,
            output_capture=output_capture,
            event_loop=event_loop,
        )

        _validate_plan_with_context(execution_context, execution_plan)

        yield execution_context
        yield from resources_manager.generate_teardown_events()

    finally:
        try:
            # If are running in another active event loop than this cleanup
            # will throw RuntimeError('Cannot run the event loop while another loop is running')
            # There is no public API that does not throw to check for this condition, so just run
            # the cleanup and ignore the exception.
            event_loop.run_until_complete(event_loop.shutdown_asyncgens())
        except RuntimeError:
            pass

        event_loop.close()


class PlanOrchestrationContextManager(ExecutionContextManager[PlanOrchestrationContext]):
    def __init__(
        self,
        context_event_generator: Callable[
            ...,
            Iterator[Union[DagsterEvent, PlanOrchestrationContext]],
        ],
        job: IJob,
        execution_plan: ExecutionPlan,
        run_config: Mapping[str, object],
        dagster_run: DagsterRun,
        instance: DagsterInstance,
        raise_on_error: Optional[bool] = False,
        output_capture: Optional[dict["StepOutputHandle", Any]] = None,
        executor_defs: Optional[Sequence[ExecutorDefinition]] = None,
        resume_from_failure=False,
    ):
        event_generator = context_event_generator(
            job,
            execution_plan,
            run_config,
            dagster_run,
            instance,
            raise_on_error,
            executor_defs,
            output_capture,
            resume_from_failure=resume_from_failure,
        )
        super().__init__(event_generator)

    @property
    def context_type(self) -> type[PlanOrchestrationContext]:
        return PlanOrchestrationContext


def orchestration_context_event_generator(
    job: IJob,
    execution_plan: ExecutionPlan,
    run_config: Mapping[str, object],
    dagster_run: DagsterRun,
    instance: DagsterInstance,
    raise_on_error: bool,
    executor_defs: Optional[Sequence[ExecutorDefinition]],
    output_capture: Optional[dict["StepOutputHandle", Any]],
    resume_from_failure: bool = False,
) -> Iterator[Union[DagsterEvent, PlanOrchestrationContext]]:
    check.invariant(executor_defs is None)
    context_creation_data = create_context_creation_data(
        job,
        execution_plan,
        run_config,
        dagster_run,
        instance,
    )

    log_manager = create_log_manager(context_creation_data)

    try:
        executor = create_executor(context_creation_data)

        execution_context = PlanOrchestrationContext(
            plan_data=create_plan_data(context_creation_data, raise_on_error, executor.retries),
            log_manager=log_manager,
            executor=executor,
            output_capture=output_capture,
            resume_from_failure=resume_from_failure,
        )

        _validate_plan_with_context(execution_context, execution_plan)

        yield execution_context
    except DagsterError as dagster_error:
        dagster_error = cast(DagsterUserCodeExecutionError, dagster_error)
        user_facing_exc_info = (
            # pylint does not know original_exc_info exists is is_user_code_error is true
            dagster_error.original_exc_info if dagster_error.is_user_code_error else sys.exc_info()
        )
        error_info = serializable_error_info_from_exc_info(user_facing_exc_info)

        event = DagsterEvent.job_failure(
            job_context_or_name=dagster_run.job_name,
            context_msg=(
                "Failure during initialization for job"
                f' "{dagster_run.job_name}". This may be due to a failure in initializing the'
                " executor or one of the loggers."
            ),
            failure_reason=RunFailureReason.JOB_INITIALIZATION_FAILURE,
            error_info=error_info,
        )
        log_manager.log_dagster_event(
            level=logging.ERROR, msg=event.message or "", dagster_event=event
        )
        yield event

        if raise_on_error:
            raise dagster_error


class PlanExecutionContextManager(ExecutionContextManager[PlanExecutionContext]):
    def __init__(
        self,
        job: IJob,
        execution_plan: ExecutionPlan,
        run_config: Mapping[str, object],
        dagster_run: DagsterRun,
        instance: DagsterInstance,
        retry_mode: RetryMode,
        scoped_resources_builder_cm: Optional[
            Callable[..., EventGenerationManager[ScopedResourcesBuilder]]
        ] = None,
        raise_on_error: Optional[bool] = False,
        output_capture: Optional[dict["StepOutputHandle", Any]] = None,
    ):
        super().__init__(
            execution_context_event_generator(
                job,
                execution_plan,
                run_config,
                dagster_run,
                instance,
                retry_mode,
                scoped_resources_builder_cm,
                raise_on_error=raise_on_error,
                output_capture=output_capture,
            )
        )

    @property
    def context_type(self) -> type[PlanExecutionContext]:
        return PlanExecutionContext


# perform any plan validation that is dependent on access to the pipeline context
def _validate_plan_with_context(job_context: IPlanContext, execution_plan: ExecutionPlan) -> None:
    validate_reexecution_memoization(job_context, execution_plan)


def create_executor(context_creation_data: ContextCreationData) -> "Executor":
    check.inst_param(context_creation_data, "context_creation_data", ContextCreationData)
    init_context = InitExecutorContext(
        job=context_creation_data.job,
        executor_def=context_creation_data.executor_def,
        executor_config=context_creation_data.resolved_run_config.execution.execution_engine_config,
        instance=context_creation_data.instance,
    )
    check_cross_process_constraints(init_context)
    creation_fn = check.not_none(context_creation_data.executor_def.executor_creation_fn)
    return creation_fn(init_context)


@contextmanager
def scoped_job_context(
    execution_plan: ExecutionPlan,
    job: IJob,
    run_config: Mapping[str, object],
    dagster_run: DagsterRun,
    instance: DagsterInstance,
    scoped_resources_builder_cm: Callable[
        ..., EventGenerationManager[ScopedResourcesBuilder]
    ] = resource_initialization_manager,
    raise_on_error: Optional[bool] = False,
) -> Generator[PlanExecutionContext, None, None]:
    """Utility context manager which acts as a very thin wrapper around
    `pipeline_initialization_manager`, iterating through all the setup/teardown events and
    discarding them.  It yields the resulting `pipeline_context`.

    Should only be used where we need to reconstruct the pipeline context, ignoring any yielded
    events (e.g. JobExecutionResult, dagstermill, unit tests, etc)
    """
    check.inst_param(execution_plan, "execution_plan", ExecutionPlan)
    check.inst_param(job, "job", IJob)
    check.mapping_param(run_config, "run_config", key_type=str)
    check.inst_param(dagster_run, "dagster_run", DagsterRun)
    check.inst_param(instance, "instance", DagsterInstance)
    check.callable_param(scoped_resources_builder_cm, "scoped_resources_builder_cm")

    initialization_manager = PlanExecutionContextManager(
        job,
        execution_plan,
        run_config,
        dagster_run,
        instance,
        RetryMode.DISABLED,
        scoped_resources_builder_cm=scoped_resources_builder_cm,
        raise_on_error=raise_on_error,
    )
    for _ in initialization_manager.prepare_context():
        pass

    try:
        yield check.inst(initialization_manager.get_context(), PlanExecutionContext)
    finally:
        for _ in initialization_manager.shutdown_context():
            pass


def create_log_manager(
    context_creation_data: ContextCreationData,
) -> DagsterLogManager:
    check.inst_param(context_creation_data, "context_creation_data", ContextCreationData)

    job_def, resolved_run_config, dagster_run = (
        context_creation_data.job_def,
        context_creation_data.resolved_run_config,
        context_creation_data.dagster_run,
    )

    # The following logic is tightly coupled to the processing of logger config in
    # python_modules/dagster/dagster/_core/system_config/objects.py#config_map_loggers
    # Changes here should be accompanied checked against that function, which applies config mapping
    # via ConfigurableDefinition (@configured) to incoming logger configs. See docstring for more details.

    loggers = []
    for logger_key, logger_def in job_def.loggers.items() or default_loggers().items():
        if logger_key in resolved_run_config.loggers:
            loggers.append(
                logger_def.logger_fn(
                    InitLoggerContext(
                        resolved_run_config.loggers.get(logger_key, {}).get("config"),
                        logger_def,
                        job_def=job_def,
                        run_id=dagster_run.run_id,
                    )
                )
            )

    if not loggers:
        for logger_def, logger_config in default_system_loggers(context_creation_data.instance):
            loggers.append(
                logger_def.logger_fn(
                    InitLoggerContext(
                        logger_config,
                        logger_def,
                        job_def=job_def,
                        run_id=dagster_run.run_id,
                    )
                )
            )

    return DagsterLogManager.create(
        loggers=loggers, dagster_run=dagster_run, instance=context_creation_data.instance
    )


def create_context_free_log_manager(
    instance: DagsterInstance, dagster_run: DagsterRun
) -> DagsterLogManager:
    """In the event of pipeline initialization failure, we want to be able to log the failure
    without a dependency on the PlanExecutionContext to initialize DagsterLogManager.

    Args:
        dagster_run (PipelineRun)
        pipeline_def (JobDefinition)
    """
    check.inst_param(instance, "instance", DagsterInstance)
    check.inst_param(dagster_run, "dagster_run", DagsterRun)

    loggers = []
    # Use the default logger
    for logger_def, logger_config in default_system_loggers(instance):
        loggers += [
            logger_def.logger_fn(
                InitLoggerContext(
                    logger_config,
                    logger_def,
                    job_def=None,
                    run_id=dagster_run.run_id,
                )
            )
        ]

    return DagsterLogManager.create(loggers=loggers, instance=instance, dagster_run=dagster_run)



import textwrap
from typing import Optional

from dagster_shared.serdes.objects import (
    ComponentTypeSnap,
    LibraryObjectKey,
    LibraryObjectSnap,
    ScaffolderSnap,
)

from dagster.components.component.component import Component
from dagster.components.scaffold.scaffold import Scaffolder, get_scaffolder


def _clean_docstring(docstring: str) -> str:
    lines = docstring.strip().splitlines()
    first_line = lines[0]
    if len(lines) == 1:
        return first_line
    else:
        rest = textwrap.dedent("\n".join(lines[1:]))
        return f"{first_line}\n{rest}"


def _get_scaffolder_snap(obj: object) -> Optional[ScaffolderSnap]:
    scaffolder = get_scaffolder(obj) if isinstance(obj, type) else None
    if not isinstance(scaffolder, Scaffolder):
        return None
    scaffolder_schema = scaffolder.get_scaffold_params()
    return ScaffolderSnap(
        schema=scaffolder_schema.model_json_schema() if scaffolder_schema else None
    )


def _get_summary_and_description(obj: object) -> tuple[Optional[str], Optional[str]]:
    docstring = obj.__doc__
    clean_docstring = _clean_docstring(docstring) if docstring else None
    summary = clean_docstring.split("\n\n")[0] if clean_docstring else None
    description = clean_docstring if clean_docstring else None
    return summary, description


def _get_component_type_snap(key: LibraryObjectKey, obj: type[Component]) -> ComponentTypeSnap:
    summary, description = _get_summary_and_description(obj)
    component_schema = obj.get_schema()
    return ComponentTypeSnap(
        key=key,
        summary=summary,
        description=description,
        schema=component_schema.model_json_schema() if component_schema else None,
        scaffolder=_get_scaffolder_snap(obj),
    )


def _get_library_object_snap(key: LibraryObjectKey, obj: object) -> LibraryObjectSnap:
    summary, description = _get_summary_and_description(obj)
    return LibraryObjectSnap(
        key=key, summary=summary, description=description, scaffolder=_get_scaffolder_snap(obj)
    )


def get_library_object_snap(key: LibraryObjectKey, obj: object) -> LibraryObjectSnap:
    if isinstance(obj, type) and issubclass(obj, Component):
        return _get_component_type_snap(key, obj)
    else:
        return _get_library_object_snap(key, obj)



import os
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, Any, NamedTuple, Optional

import yaml

import dagster._check as check
from dagster._core.instance.config import DAGSTER_CONFIG_YAML_FILENAME, dagster_instance_config
from dagster._serdes import ConfigurableClassData, class_from_code_pointer, whitelist_for_serdes

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance, DagsterInstanceOverrides
    from dagster._core.launcher.base import RunLauncher
    from dagster._core.run_coordinator.base import RunCoordinator
    from dagster._core.scheduler.scheduler import Scheduler
    from dagster._core.secrets.loader import SecretsLoader
    from dagster._core.storage.base_storage import DagsterStorage
    from dagster._core.storage.compute_log_manager import ComputeLogManager
    from dagster._core.storage.event_log.base import EventLogStorage
    from dagster._core.storage.root import LocalArtifactStorage
    from dagster._core.storage.runs.base import RunStorage
    from dagster._core.storage.schedules.base import ScheduleStorage


def compute_logs_directory(base: str) -> str:
    return os.path.join(base, "storage")


def _runs_directory(base: str) -> str:
    return os.path.join(base, "history", "")


def _event_logs_directory(base: str) -> str:
    return os.path.join(base, "history", "runs", "")


def _schedule_directory(base: str) -> str:
    return os.path.join(base, "schedules")


def configurable_class_data(config_field: Mapping[str, Any]) -> ConfigurableClassData:
    return ConfigurableClassData(
        check.str_elem(config_field, "module"),
        check.str_elem(config_field, "class"),
        yaml.dump(check.opt_dict_elem(config_field, "config"), default_flow_style=False),
    )


def configurable_class_data_or_default(
    config_value: Mapping[str, Any], field_name: str, default: Optional[ConfigurableClassData]
) -> Optional[ConfigurableClassData]:
    return (
        configurable_class_data(config_value[field_name])
        if config_value.get(field_name)
        else default
    )


def configurable_secrets_loader_data(
    config_field: Mapping[str, Any], default: Optional[ConfigurableClassData]
) -> Optional[ConfigurableClassData]:
    if not config_field:
        return default
    elif "custom" in config_field:
        return configurable_class_data(config_field["custom"])
    else:
        return None


def configurable_storage_data(
    config_field: Mapping[str, Any], defaults: Mapping[str, Optional[ConfigurableClassData]]
) -> Sequence[Optional[ConfigurableClassData]]:
    storage_data: ConfigurableClassData
    run_storage_data: Optional[ConfigurableClassData]
    event_storage_data: Optional[ConfigurableClassData]
    schedule_storage_data: Optional[ConfigurableClassData]

    if not config_field:
        storage_data = check.not_none(defaults.get("storage"))
        run_storage_data = check.not_none(defaults.get("run_storage"))
        event_storage_data = check.not_none(defaults.get("event_log_storage"))
        schedule_storage_data = check.not_none(defaults.get("schedule_storage"))
    elif "postgres" in config_field:
        config_yaml = yaml.dump(config_field["postgres"], default_flow_style=False)
        storage_data = ConfigurableClassData(
            module_name="dagster_postgres",
            class_name="DagsterPostgresStorage",
            config_yaml=config_yaml,
        )
        # for backwards compatibility
        run_storage_data = ConfigurableClassData(
            module_name="dagster_postgres",
            class_name="PostgresRunStorage",
            config_yaml=config_yaml,
        )
        event_storage_data = ConfigurableClassData(
            module_name="dagster_postgres",
            class_name="PostgresEventLogStorage",
            config_yaml=config_yaml,
        )
        schedule_storage_data = ConfigurableClassData(
            module_name="dagster_postgres",
            class_name="PostgresScheduleStorage",
            config_yaml=config_yaml,
        )

    elif "mysql" in config_field:
        config_yaml = yaml.dump(config_field["mysql"], default_flow_style=False)
        storage_data = ConfigurableClassData(
            module_name="dagster_mysql",
            class_name="DagsterMySQLStorage",
            config_yaml=config_yaml,
        )
        # for backwards compatibility
        run_storage_data = ConfigurableClassData(
            module_name="dagster_mysql",
            class_name="MySQLRunStorage",
            config_yaml=config_yaml,
        )
        event_storage_data = ConfigurableClassData(
            module_name="dagster_mysql",
            class_name="MySQLEventLogStorage",
            config_yaml=config_yaml,
        )
        schedule_storage_data = ConfigurableClassData(
            module_name="dagster_mysql",
            class_name="MySQLScheduleStorage",
            config_yaml=config_yaml,
        )

    elif "sqlite" in config_field:
        base_dir = config_field["sqlite"]["base_dir"]
        storage_data = ConfigurableClassData(
            "dagster._core.storage.sqlite_storage",
            "DagsterSqliteStorage",
            yaml.dump({"base_dir": base_dir}, default_flow_style=False),
        )

        # Back-compat fo the legacy storage field only works if the base_dir is a string
        # (env var doesn't work since each storage has a different value for the base_dir field)
        if isinstance(base_dir, str):
            run_storage_data = ConfigurableClassData(
                "dagster._core.storage.runs",
                "SqliteRunStorage",
                yaml.dump({"base_dir": _runs_directory(base_dir)}, default_flow_style=False),
            )

            event_storage_data = ConfigurableClassData(
                "dagster._core.storage.event_log",
                "SqliteEventLogStorage",
                yaml.dump({"base_dir": _event_logs_directory(base_dir)}, default_flow_style=False),
            )

            schedule_storage_data = ConfigurableClassData(
                "dagster._core.storage.schedules",
                "SqliteScheduleStorage",
                yaml.dump({"base_dir": _schedule_directory(base_dir)}, default_flow_style=False),
            )
        else:
            run_storage_data = None
            event_storage_data = None
            schedule_storage_data = None
    else:
        storage_data = configurable_class_data(config_field["custom"])
        storage_config_yaml = yaml.dump(
            {
                "module_name": storage_data.module_name,
                "class_name": storage_data.class_name,
                "config_yaml": storage_data.config_yaml,
            },
            default_flow_style=False,
        )
        run_storage_data = ConfigurableClassData(
            "dagster._core.storage.legacy_storage", "LegacyRunStorage", storage_config_yaml
        )
        event_storage_data = ConfigurableClassData(
            "dagster._core.storage.legacy_storage", "LegacyEventLogStorage", storage_config_yaml
        )
        schedule_storage_data = ConfigurableClassData(
            "dagster._core.storage.legacy_storage", "LegacyScheduleStorage", storage_config_yaml
        )

    return [storage_data, run_storage_data, event_storage_data, schedule_storage_data]


@whitelist_for_serdes
class InstanceRef(
    NamedTuple(
        "_InstanceRef",
        [
            ("local_artifact_storage_data", ConfigurableClassData),
            ("compute_logs_data", ConfigurableClassData),
            ("scheduler_data", Optional[ConfigurableClassData]),
            ("run_coordinator_data", Optional[ConfigurableClassData]),
            ("run_launcher_data", Optional[ConfigurableClassData]),
            ("settings", Mapping[str, object]),
            # Required for backwards compatibility, but going forward will be unused by new versions
            # of DagsterInstance, which instead will instead grab the constituent storages from the
            # unified `storage_data`, if it is populated.
            ("run_storage_data", Optional[ConfigurableClassData]),
            ("event_storage_data", Optional[ConfigurableClassData]),
            ("schedule_storage_data", Optional[ConfigurableClassData]),
            ("custom_instance_class_data", Optional[ConfigurableClassData]),
            # unified storage field
            ("storage_data", Optional[ConfigurableClassData]),
            ("secrets_loader_data", Optional[ConfigurableClassData]),
        ],
    )
):
    """Serializable representation of a :py:class:`DagsterInstance`.

    Users should not instantiate this class directly.
    """

    def __new__(
        cls,
        local_artifact_storage_data: ConfigurableClassData,
        compute_logs_data: ConfigurableClassData,
        scheduler_data: Optional[ConfigurableClassData],
        run_coordinator_data: Optional[ConfigurableClassData],
        run_launcher_data: Optional[ConfigurableClassData],
        settings: Mapping[str, object],
        run_storage_data: Optional[ConfigurableClassData],
        event_storage_data: Optional[ConfigurableClassData],
        schedule_storage_data: Optional[ConfigurableClassData],
        custom_instance_class_data: Optional[ConfigurableClassData] = None,
        storage_data: Optional[ConfigurableClassData] = None,
        secrets_loader_data: Optional[ConfigurableClassData] = None,
    ):
        return super(cls, InstanceRef).__new__(
            cls,
            local_artifact_storage_data=check.inst_param(
                local_artifact_storage_data, "local_artifact_storage_data", ConfigurableClassData
            ),
            compute_logs_data=check.inst_param(
                compute_logs_data, "compute_logs_data", ConfigurableClassData
            ),
            scheduler_data=check.opt_inst_param(
                scheduler_data, "scheduler_data", ConfigurableClassData
            ),
            run_coordinator_data=check.opt_inst_param(
                run_coordinator_data, "run_coordinator_data", ConfigurableClassData
            ),
            run_launcher_data=check.opt_inst_param(
                run_launcher_data, "run_launcher_data", ConfigurableClassData
            ),
            settings=check.opt_mapping_param(settings, "settings", key_type=str),
            run_storage_data=check.opt_inst_param(
                run_storage_data, "run_storage_data", ConfigurableClassData
            ),
            event_storage_data=check.opt_inst_param(
                event_storage_data, "event_storage_data", ConfigurableClassData
            ),
            schedule_storage_data=check.opt_inst_param(
                schedule_storage_data, "schedule_storage_data", ConfigurableClassData
            ),
            custom_instance_class_data=check.opt_inst_param(
                custom_instance_class_data,
                "instance_class",
                ConfigurableClassData,
            ),
            storage_data=check.opt_inst_param(storage_data, "storage_data", ConfigurableClassData),
            secrets_loader_data=check.opt_inst_param(
                secrets_loader_data, "secrets_loader_data", ConfigurableClassData
            ),
        )

    @staticmethod
    def config_defaults(base_dir: str) -> Mapping[str, Optional[ConfigurableClassData]]:
        default_run_storage_data = ConfigurableClassData(
            "dagster._core.storage.runs",
            "SqliteRunStorage",
            yaml.dump({"base_dir": _runs_directory(base_dir)}, default_flow_style=False),
        )
        default_event_log_storage_data = ConfigurableClassData(
            "dagster._core.storage.event_log",
            "SqliteEventLogStorage",
            yaml.dump({"base_dir": _event_logs_directory(base_dir)}, default_flow_style=False),
        )
        default_schedule_storage_data = ConfigurableClassData(
            "dagster._core.storage.schedules",
            "SqliteScheduleStorage",
            yaml.dump({"base_dir": _schedule_directory(base_dir)}, default_flow_style=False),
        )

        return {
            "local_artifact_storage": ConfigurableClassData(
                "dagster._core.storage.root",
                "LocalArtifactStorage",
                yaml.dump({"base_dir": base_dir}, default_flow_style=False),
            ),
            "storage": ConfigurableClassData(
                "dagster._core.storage.sqlite_storage",
                "DagsterSqliteStorage",
                yaml.dump({"base_dir": base_dir}, default_flow_style=False),
            ),
            "compute_logs": ConfigurableClassData(
                "dagster._core.storage.local_compute_log_manager",
                "LocalComputeLogManager",
                yaml.dump({"base_dir": compute_logs_directory(base_dir)}, default_flow_style=False),
            ),
            "scheduler": ConfigurableClassData(
                "dagster._core.scheduler",
                "DagsterDaemonScheduler",
                yaml.dump({}),
            ),
            "run_coordinator": ConfigurableClassData(
                "dagster.core.run_coordinator",
                "QueuedRunCoordinator",
                yaml.dump({}),
            ),
            "run_launcher": ConfigurableClassData(
                "dagster",
                "DefaultRunLauncher",
                yaml.dump({}),
            ),
            # For back-compat, the default is actually set in the secrets_loader property above,
            # so that old clients loading new config don't try to load a class that they
            # don't recognize
            "secrets": None,
            # LEGACY DEFAULTS
            "run_storage": default_run_storage_data,
            "event_log_storage": default_event_log_storage_data,
            "schedule_storage": default_schedule_storage_data,
        }

    @staticmethod
    def from_dir(
        base_dir: str,
        *,
        config_dir: Optional[str] = None,
        config_filename: str = DAGSTER_CONFIG_YAML_FILENAME,
        overrides: Optional["DagsterInstanceOverrides"] = None,
    ) -> "InstanceRef":
        if config_dir is None:
            config_dir = base_dir

        overrides = check.opt_mapping_param(overrides, "overrides")
        config_value, custom_instance_class = dagster_instance_config(
            config_dir, config_filename=config_filename, overrides=overrides
        )

        if custom_instance_class:
            config_keys = set(custom_instance_class.config_schema().keys())  # type: ignore  # (undefined method)
            custom_instance_class_config = {
                key: val for key, val in config_value.items() if key in config_keys
            }
            custom_instance_class_data = ConfigurableClassData(
                config_value["instance_class"]["module"],
                config_value["instance_class"]["class"],
                yaml.dump(custom_instance_class_config, default_flow_style=False),
            )
            defaults = custom_instance_class.config_defaults(base_dir)  # type: ignore  # (undefined method)
        else:
            custom_instance_class_data = None
            defaults = InstanceRef.config_defaults(base_dir)

        local_artifact_storage_data = configurable_class_data_or_default(
            config_value, "local_artifact_storage", defaults["local_artifact_storage"]
        )

        compute_logs_data = configurable_class_data_or_default(
            config_value,
            "compute_logs",
            defaults["compute_logs"],
        )

        if (
            config_value.get("run_storage")
            or config_value.get("event_log_storage")
            or config_value.get("schedule_storage")
        ):
            # using legacy config, specifying config for each of the constituent storages, make sure
            # to create a composite storage
            run_storage_data = configurable_class_data_or_default(
                config_value, "run_storage", defaults["run_storage"]
            )
            event_storage_data = configurable_class_data_or_default(
                config_value, "event_log_storage", defaults["event_log_storage"]
            )
            schedule_storage_data = configurable_class_data_or_default(
                config_value, "schedule_storage", defaults["schedule_storage"]
            )
            storage_data = ConfigurableClassData(
                module_name="dagster._core.storage.legacy_storage",
                class_name="CompositeStorage",
                config_yaml=yaml.dump(
                    {
                        "run_storage": {
                            "module_name": run_storage_data.module_name,  # type: ignore  # (possible none)
                            "class_name": run_storage_data.class_name,  # type: ignore  # (possible none)
                            "config_yaml": run_storage_data.config_yaml,  # type: ignore  # (possible none)
                        },
                        "event_log_storage": {
                            "module_name": event_storage_data.module_name,  # type: ignore  # (possible none)
                            "class_name": event_storage_data.class_name,  # type: ignore  # (possible none)
                            "config_yaml": event_storage_data.config_yaml,  # type: ignore  # (possible none)
                        },
                        "schedule_storage": {
                            "module_name": schedule_storage_data.module_name,  # type: ignore  # (possible none)
                            "class_name": schedule_storage_data.class_name,  # type: ignore  # (possible none)
                            "config_yaml": schedule_storage_data.config_yaml,  # type: ignore  # (possible none)
                        },
                    },
                    default_flow_style=False,
                ),
            )

        else:
            [
                storage_data,
                run_storage_data,
                event_storage_data,
                schedule_storage_data,
            ] = configurable_storage_data(
                config_value.get("storage"),  # type: ignore  # (possible none)
                defaults,
            )

        scheduler_data = configurable_class_data_or_default(
            config_value, "scheduler", defaults["scheduler"]
        )

        if config_value.get("run_queue"):
            run_coordinator_data = configurable_class_data(
                {
                    "module": "dagster.core.run_coordinator",
                    "class": "QueuedRunCoordinator",
                    "config": config_value["run_queue"],
                }
            )
        else:
            run_coordinator_data = configurable_class_data_or_default(
                config_value,
                "run_coordinator",
                defaults["run_coordinator"],
            )

        run_launcher_data = configurable_class_data_or_default(
            config_value,
            "run_launcher",
            defaults["run_launcher"],
        )

        secrets_loader_data = configurable_secrets_loader_data(
            config_value.get("secrets"),  # type: ignore  # (possible none)
            defaults["secrets"],
        )

        settings_keys = {
            "telemetry",
            "python_logs",
            "run_monitoring",
            "run_retries",
            "code_servers",
            "retention",
            "backfills",
            "sensors",
            "schedules",
            "nux",
            "auto_materialize",
            "concurrency",
        }
        settings = {key: config_value.get(key) for key in settings_keys if config_value.get(key)}

        return InstanceRef(
            local_artifact_storage_data=local_artifact_storage_data,  # type: ignore  # (possible none)
            run_storage_data=run_storage_data,
            event_storage_data=event_storage_data,
            compute_logs_data=compute_logs_data,  # type: ignore  # (possible none)
            schedule_storage_data=schedule_storage_data,
            scheduler_data=scheduler_data,
            run_coordinator_data=run_coordinator_data,
            run_launcher_data=run_launcher_data,
            settings=settings,
            custom_instance_class_data=custom_instance_class_data,
            storage_data=storage_data,
            secrets_loader_data=secrets_loader_data,
        )

    @staticmethod
    def from_dict(instance_ref_dict):
        def value_for_ref_item(k, v):
            if v is None:
                return None
            if k == "settings":
                return v
            return ConfigurableClassData(*v)

        return InstanceRef(**{k: value_for_ref_item(k, v) for k, v in instance_ref_dict.items()})  # pyright: ignore[reportArgumentType]

    @property
    def local_artifact_storage(self) -> "LocalArtifactStorage":
        from dagster._core.storage.root import LocalArtifactStorage

        return self.local_artifact_storage_data.rehydrate(as_type=LocalArtifactStorage)

    @property
    def storage(self) -> Optional["DagsterStorage"]:
        from dagster._core.storage.base_storage import DagsterStorage

        return self.storage_data.rehydrate(as_type=DagsterStorage) if self.storage_data else None

    @property
    def run_storage(self) -> Optional["RunStorage"]:
        from dagster._core.storage.runs.base import RunStorage

        return (
            self.run_storage_data.rehydrate(as_type=RunStorage) if self.run_storage_data else None
        )

    @property
    def event_storage(self) -> Optional["EventLogStorage"]:
        from dagster._core.storage.event_log.base import EventLogStorage

        return (
            self.event_storage_data.rehydrate(as_type=EventLogStorage)
            if self.event_storage_data
            else None
        )

    @property
    def schedule_storage(self) -> Optional["ScheduleStorage"]:
        from dagster._core.storage.schedules.base import ScheduleStorage

        return (
            self.schedule_storage_data.rehydrate(as_type=ScheduleStorage)
            if self.schedule_storage_data
            else None
        )

    @property
    def compute_log_manager(self) -> "ComputeLogManager":
        from dagster._core.storage.compute_log_manager import ComputeLogManager

        return self.compute_logs_data.rehydrate(as_type=ComputeLogManager)

    @property
    def scheduler(self) -> Optional["Scheduler"]:
        from dagster._core.scheduler.scheduler import Scheduler

        return self.scheduler_data.rehydrate(as_type=Scheduler) if self.scheduler_data else None

    @property
    def run_coordinator(self) -> Optional["RunCoordinator"]:
        from dagster._core.run_coordinator.base import RunCoordinator

        return (
            self.run_coordinator_data.rehydrate(as_type=RunCoordinator)
            if self.run_coordinator_data
            else None
        )

    @property
    def run_launcher(self) -> Optional["RunLauncher"]:
        from dagster._core.launcher.base import RunLauncher

        return (
            self.run_launcher_data.rehydrate(as_type=RunLauncher)
            if self.run_launcher_data
            else None
        )

    @property
    def secrets_loader(self) -> Optional["SecretsLoader"]:
        from dagster._core.secrets.env_file import PerProjectEnvFileLoader
        from dagster._core.secrets.loader import SecretsLoader

        # Default PerProjectEnvFileLoader is injected here, rather than
        # in config_defaults, to avoid back-compat issues when loading the
        # config on older versions where PerProjectEnvFileLoader was not
        # defined.
        return (
            self.secrets_loader_data.rehydrate(as_type=SecretsLoader)
            if self.secrets_loader_data
            else PerProjectEnvFileLoader()
        )

    @property
    def custom_instance_class(self) -> type["DagsterInstance"]:
        return (  # type: ignore  # (ambiguous return type)
            class_from_code_pointer(
                self.custom_instance_class_data.module_name,
                self.custom_instance_class_data.class_name,
            )
            if self.custom_instance_class_data
            else None
        )

    @property
    def custom_instance_class_config(self) -> Mapping[str, Any]:
        return (
            self.custom_instance_class_data.config_dict if self.custom_instance_class_data else {}
        )

    def to_dict(self) -> Mapping[str, Any]:
        return self._asdict()



import importlib
from pathlib import Path
from types import ModuleType

from dagster._annotations import deprecated
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._utils.warnings import suppress_dagster_warnings
from dagster.components.core.context import ComponentLoadContext, use_component_load_context


@deprecated(breaking_version="0.2.0")
@suppress_dagster_warnings
def build_component_defs(components_root: Path) -> Definitions:
    """Build a Definitions object for all the component instances in a given code location.

    Args:
        components_root (Path): The path to the components root. This is a directory containing
            subdirectories with component instances.
    """
    defs_root = importlib.import_module(
        f"{Path(components_root).parent.name}.{Path(components_root).name}"
    )

    return load_defs(defs_root=defs_root)


# Public method so optional Nones are fine
@suppress_dagster_warnings
def load_defs(defs_root: ModuleType) -> Definitions:
    """Constructs a Definitions object, loading all Dagster defs in the given module.

    Args:
        defs_root (Path): The path to the defs root, typically `package.defs`.
        resources (Optional[Mapping[str, object]]): A mapping of resource keys to resources
            to apply to the definitions.
    """
    from dagster.components.core.defs_module import DefsModuleComponent

    # create a top-level DefsModule component from the root module
    context = ComponentLoadContext.for_module(defs_root)
    root_component = DefsModuleComponent.from_context(context)
    if root_component is None:
        raise DagsterInvalidDefinitionError("Could not resolve root module to a component.")

    with use_component_load_context(context):
        return root_component.build_defs(context)



from dagster.components.component.component import Component as Component
from dagster.components.component.component_loader import component as component
from dagster.components.component.component_scaffolder import (
    DefaultComponentScaffolder as DefaultComponentScaffolder,
)
from dagster.components.component_scaffolding import scaffold_component as scaffold_component
from dagster.components.components import (
    DefinitionsComponent as DefinitionsComponent,
    DefsFolderComponent as DefsFolderComponent,
    PipesSubprocessScriptCollectionComponent as PipesSubprocessScriptCollectionComponent,
)
from dagster.components.core.context import ComponentLoadContext as ComponentLoadContext
from dagster.components.core.load_defs import (
    build_component_defs as build_component_defs,
    load_defs as load_defs,
)
from dagster.components.resolved.base import Resolvable as Resolvable
from dagster.components.resolved.context import ResolutionContext as ResolutionContext
from dagster.components.resolved.core_models import (
    AssetAttributesModel as AssetAttributesModel,
    AssetPostProcessorModel as AssetPostProcessorModel,
    ResolvedAssetSpec as ResolvedAssetSpec,
)
from dagster.components.resolved.metadata import ResolvableFieldInfo as ResolvableFieldInfo
from dagster.components.resolved.model import (
    Injectable as Injectable,
    Injected as Injected,
    Model as Model,
    Resolver as Resolver,
)
from dagster.components.scaffold.scaffold import (
    Scaffolder as Scaffolder,
    ScaffolderUnavailableReason as ScaffolderUnavailableReason,
    ScaffoldRequest as ScaffoldRequest,
)



from typing import Any

from dagster.components.scaffold.scaffold import Scaffolder, ScaffoldRequest


class DefaultComponentScaffolder(Scaffolder):
    def scaffold(self, request: ScaffoldRequest, params: Any) -> None:
        # This will be deleted once all components are converted to the new ComponentScaffolder API
        from dagster.components.component_scaffolding import scaffold_component

        scaffold_component(request, params.model_dump() if params else {})



from typing import TYPE_CHECKING, Any, Callable, TypeVar

if TYPE_CHECKING:
    from dagster.components.component.component import Component
    from dagster.components.core.context import ComponentLoadContext

COMPONENT_LOADER_FN_ATTR = "__dagster_component_loader_fn"


T_Component = TypeVar("T_Component", bound="Component")


def component(
    fn: Callable[["ComponentLoadContext"], T_Component],
) -> Callable[["ComponentLoadContext"], T_Component]:
    setattr(fn, COMPONENT_LOADER_FN_ATTR, True)
    return fn


def is_component_loader(obj: Any) -> bool:
    return getattr(obj, COMPONENT_LOADER_FN_ATTR, False)



from dagster._core.definitions.decorators.asset_decorator import asset
from dagster.components.lib.shim_components.base import ShimScaffolder
from dagster.components.scaffold.scaffold import scaffold_with


class AssetScaffolder(ShimScaffolder):
    def get_text(self, filename: str) -> str:
        return f"""# import dagster as dg
# 
#
# @dg.asset
# def {filename}(context: dg.AssetExecutionContext) -> dg.MaterializeResult: ...

"""


scaffold_with(AssetScaffolder)(asset)



import inspect
from collections.abc import Mapping, Sequence
from dataclasses import MISSING, fields, is_dataclass
from enum import Enum, auto
from functools import partial
from types import GenericAlias
from typing import Annotated, Any, Final, Literal, Optional, TypeVar, Union, get_args, get_origin

from dagster_shared.record import get_record_annotations, get_record_defaults, is_record
from dagster_shared.yaml_utils import parse_yaml_with_source_positions
from pydantic import BaseModel, PydanticSchemaGenerationError, create_model
from pydantic.fields import Field, FieldInfo
from typing_extensions import TypeGuard

from dagster import _check as check
from dagster._utils.pydantic_yaml import _parse_and_populate_model_with_annotated_errors
from dagster.components.resolved.context import ResolutionContext
from dagster.components.resolved.errors import ResolutionException
from dagster.components.resolved.model import Model, Resolver

try:
    # this type only exists in python 3.10+
    from types import UnionType  # type: ignore
except ImportError:
    UnionType = Union


class _TypeContainer(Enum):
    SEQUENCE = auto()
    OPTIONAL = auto()


_DERIVED_MODEL_REGISTRY = {}


class Resolvable:
    """This base class makes something able to "resolve" from yaml.

    This is done by:
    1) Deriving a pydantic model to provide as schema for the yaml.
    2) Resolving an instance of the class by recursing over an instance
        of the derived model loaded from schema compliant yaml and
        evaluating any template strings.

    The fields/__init__ arguments of the class can be Annotated with
    Resolver to customize the resolution or model derivation.
    """

    @classmethod
    def model(cls) -> type[BaseModel]:
        return derive_model_type(cls)

    @classmethod
    def resolve_from_model(cls, context: "ResolutionContext", model: BaseModel):
        return cls(**resolve_fields(model, cls, context))

    @classmethod
    def resolve_from_yaml(cls, yaml: str):
        parsed_and_src_tree = parse_yaml_with_source_positions(yaml)
        model_cls = cls.model()
        if parsed_and_src_tree:
            model = _parse_and_populate_model_with_annotated_errors(
                cls=model_cls,
                obj_parse_root=parsed_and_src_tree,
                obj_key_path_prefix=[],
            )
        else:  # yaml parsed as None
            model = model_cls()
        # support adding scopes?
        context = ResolutionContext.default(
            parsed_and_src_tree.source_position_tree if parsed_and_src_tree else None
        )
        return cls.resolve_from_model(context, model)


# marker type for skipping kwargs and triggering defaults
# must be a string to make sure it is json serializable
_Unset: Final[str] = "__DAGSTER_UNSET_DEFAULT__"


def derive_model_type(
    target_type: type[Resolvable],
) -> type[BaseModel]:
    if target_type not in _DERIVED_MODEL_REGISTRY:
        model_name = f"{target_type.__name__}Model"

        model_fields: dict[
            str, Any
        ] = {}  # use Any to appease type checker when **-ing in to create_model

        for name, (annotation, has_default) in _get_annotations(target_type).items():
            field_resolver = _get_resolver(annotation, name)
            field_name = field_resolver.model_field_name or name
            field_type = field_resolver.model_field_type or annotation

            field_infos = []

            if has_default:
                # use a marker value that will cause the kwarg
                # to get omitted when we resolve fields in order
                # to trigger the default on the target type
                field_infos.append(Field(default=_Unset))

            if field_resolver.can_inject:  # derive and serve via model_field_type
                field_type = Union[field_type, str]

            model_fields[field_name] = (
                field_type,
                FieldInfo.merge_field_infos(*field_infos),
            )

        try:
            _DERIVED_MODEL_REGISTRY[target_type] = create_model(
                model_name,
                __base__=Model,
                **model_fields,
            )
        except PydanticSchemaGenerationError as e:
            raise ResolutionException(f"Unable to derive Model for {target_type}") from e

    return _DERIVED_MODEL_REGISTRY[target_type]


def _is_implicitly_resolved_type(annotation):
    if annotation in (int, float, str, bool, Any, type(None)):
        return True

    if _safe_is_subclass(annotation, Resolvable):
        return False

    if _safe_is_subclass(annotation, BaseModel):
        return True

    origin = get_origin(annotation)
    args = get_args(annotation)

    if origin in (Union, UnionType, list, Sequence, tuple, dict, Mapping) and all(
        _is_implicitly_resolved_type(arg) for arg in args
    ):
        return True

    if origin is Literal and all(_is_implicitly_resolved_type(type(arg)) for arg in args):
        return True

    return False


def _get_annotations(resolved_type: type[Resolvable]):
    annotations: dict[str, tuple[Any, bool]] = {}
    init_kwargs = _get_init_kwargs(resolved_type)
    if is_dataclass(resolved_type):
        for f in fields(resolved_type):
            has_default = f.default is not MISSING or f.default_factory is not MISSING
            annotations[f.name] = (f.type, has_default)
        return annotations
    elif _safe_is_subclass(resolved_type, BaseModel):
        for name, field_info in resolved_type.model_fields.items():
            has_default = not field_info.is_required()
            annotations[name] = (field_info.rebuild_annotation(), has_default)
        return annotations
    elif is_record(resolved_type):
        defaults = get_record_defaults(resolved_type)
        for name, ttype in get_record_annotations(resolved_type).items():
            annotations[name] = (ttype, name in defaults)
        return annotations
    elif init_kwargs is not None:
        return init_kwargs
    # can update to support:
    # * @record
    else:
        raise ResolutionException(
            f"Invalid Resolved type {resolved_type} could not determine fields, expected:\n"
            "* class with __init__\n"
            "* @dataclass\n"
            "* pydantic Model\n"
            "* @record\n"
        )


def _get_init_kwargs(target_type: type[Resolvable]):
    if target_type.__init__ is object.__init__:
        return None

    sig = inspect.signature(target_type.__init__)
    fields = {}

    skipped_self = False
    for name, param in sig.parameters.items():
        if not skipped_self:
            skipped_self = True
            continue

        if param.kind == param.POSITIONAL_ONLY:
            raise ResolutionException(
                f"Invalid Resolved type {target_type}: __init__ contains positional only parameter."
            )
        if param.kind in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
            continue
        if param.annotation == param.empty:
            raise ResolutionException(
                f"Invalid Resolved type {target_type}: __init__ parameter {name} has no type hint."
            )

        fields[name] = (param.annotation, param.default is not param.empty)
    return fields


def resolve_fields(
    model: BaseModel,
    resolved_cls: type[Resolvable],
    context: "ResolutionContext",
) -> Mapping[str, Any]:
    """Returns a mapping of field names to resolved values for those fields."""
    field_resolvers = {
        field_name: _get_resolver(annotation, field_name)
        for field_name, (annotation, _) in _get_annotations(resolved_cls).items()
    }

    return {
        field_name: resolver.execute(context=context, model=model, field_name=field_name)
        for field_name, resolver in field_resolvers.items()
        # filter out _Unset to trigger defaults
        if getattr(model, resolver.model_field_name or field_name) is not _Unset
    }


TType = TypeVar("TType", bound=type)


def _safe_is_subclass(obj, cls: TType) -> TypeGuard[type[TType]]:
    return (
        isinstance(obj, type)
        and not isinstance(obj, GenericAlias)  # prevent exceptions on 3.9
        and issubclass(obj, cls)
    )


def _get_resolver(annotation: Any, field_name: str) -> "Resolver":
    origin = get_origin(annotation)
    args = get_args(annotation)

    # explicit field level Resolver
    if origin is Annotated:
        resolver = next((arg for arg in args if isinstance(arg, Resolver)), None)
        if resolver:
            check.invariant(
                _is_implicitly_resolved_type(args[0]) or resolver.model_field_type,
                f"Resolver for {field_name} must define model_field_type {args[0]} is not model compliant.",
            )
            return resolver

    if _is_implicitly_resolved_type(annotation):
        return Resolver.default()

    # nested or implicit
    res = _dig_for_resolver(annotation, [])
    if res:
        return res
    raise ResolutionException(
        f"Could not derive resolver for annotation {field_name}: {annotation}.\n"
        "Field types are expected to be:\n"
        "* serializable types such as str, float, int, bool, list, etc\n"
        "* pydantic Models\n"
        "* Annotated with an appropriate Resolver."
    )


def _dig_for_resolver(annotation, path: Sequence[_TypeContainer]):
    origin = get_origin(annotation)
    args = get_args(annotation)
    if _safe_is_subclass(annotation, Resolvable):
        return Resolver(
            partial(
                _resolve_at_path,
                container_path=path,
                resolver=annotation.resolve_from_model,
            ),
            model_field_type=_wrap(annotation.model(), path),
        )

    if origin is Annotated:
        resolver = next((arg for arg in args if isinstance(arg, Resolver)), None)
        if resolver:
            check.invariant(
                _is_implicitly_resolved_type(args[0]) or resolver.model_field_type,
                f"Nested resolver must define model_field_type {args[0]} is not model compliant.",
            )
            # need to ensure nested resolvers set their model type
            return Resolver(
                resolver.fn.__class__(
                    partial(
                        _resolve_at_path,
                        container_path=path,
                        resolver=resolver.fn.callable,
                    )
                ),
                model_field_type=_wrap(resolver.model_field_type or args[0], path),
            )

    if origin in (Union, UnionType) and len(args) == 2:
        left_t, right_t = args
        if right_t is type(None):
            res = _dig_for_resolver(left_t, [*path, _TypeContainer.OPTIONAL])
            if res:
                return res

    elif origin in (Sequence, tuple, list):  # should look for tuple[T, ...] specifically
        res = _dig_for_resolver(args[0], [*path, _TypeContainer.SEQUENCE])
        if res:
            return res


def _wrap(ttype, path: Sequence[_TypeContainer]):
    result_type = ttype
    for container in reversed(path):
        if container is _TypeContainer.OPTIONAL:
            result_type = Optional[result_type]
        elif container is _TypeContainer.SEQUENCE:
            # use tuple instead of Sequence for perf
            result_type = tuple[result_type, ...]
        else:
            check.assert_never(container)
    return result_type


def _resolve_at_path(
    context: "ResolutionContext",
    value: Any,
    container_path: Sequence[_TypeContainer],
    resolver,
):
    if not container_path:
        return resolver(context, value)

    container = container_path[0]
    inner_path = container_path[1:]
    if container is _TypeContainer.OPTIONAL:
        return _resolve_at_path(context, value, inner_path, resolver) if value is not None else None
    elif container is _TypeContainer.SEQUENCE:
        return [
            _resolve_at_path(context.at_path(idx), i, inner_path, resolver)
            for idx, i in enumerate(value)
        ]

    check.assert_never(container)



"""extract asset_keys index columns.

Revision ID: 7b8304b4429d
Revises: 1a2d72f6b24e
Create Date: 2021-07-06 10:52:50.862728

"""

from dagster._core.storage.migration.utils import extract_asset_keys_idx_columns

# revision identifiers, used by Alembic.
revision = "7b8304b4429d"
down_revision = "1a2d72f6b24e"
branch_labels = None
depends_on = None


def upgrade():
    extract_asset_keys_idx_columns()


def downgrade():
    pass



from pathlib import Path
from typing import Optional

import click
from dagster_shared.serdes.objects import LibraryObjectKey
from pydantic import TypeAdapter

from dagster.components.component_scaffolding import scaffold_object
from dagster.components.core.library_object import load_library_object
from dagster.components.scaffold.scaffold import ScaffolderUnavailableReason, get_scaffolder


@click.group(name="scaffold")
def scaffold_cli() -> None:
    """Commands for scaffolding Dagster components and related entities."""


@scaffold_cli.command(name="object")
@click.argument("typename", type=str)
@click.argument("path", type=Path)
@click.option("--json-params", type=str, default=None)
@click.option(
    "--scaffold-format",
    type=click.Choice(["yaml", "python"], case_sensitive=False),
    help="Format of the component configuration (yaml or python)",
)
def scaffold_object_command(
    typename: str,
    path: Path,
    json_params: Optional[str],
    scaffold_format: str,
) -> None:
    key = LibraryObjectKey.from_typename(typename)
    obj = load_library_object(key)

    if json_params:
        scaffolder = get_scaffolder(obj)
        if isinstance(scaffolder, ScaffolderUnavailableReason):
            raise Exception(
                f"Object {obj} does not have a scaffolder. Reason: {scaffolder.message}."
            )
        scaffold_params = TypeAdapter(scaffolder.get_scaffold_params()).validate_json(json_params)
    else:
        scaffold_params = {}

    scaffold_object(path, obj, typename, scaffold_params, scaffold_format)



"""Sample local components for testing validation. Paired with test cases
in integration_tests/integration_test_defs/validation.
"""

from collections.abc import Mapping
from dataclasses import dataclass
from typing import Annotated, Any

from pydantic import BaseModel, ConfigDict

from dagster._core.definitions.definitions_class import Definitions
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.resolved.base import Resolvable
from dagster.components.resolved.model import Resolver


def _inner_error():
    raise Exception("boom")


def _error():
    _inner_error()


def _maybe_throw(ctx, throw):
    if throw:
        _error()
    return throw


@dataclass
class MyComponent(Component, Resolvable):
    a_string: str
    an_int: int
    throw: Annotated[bool, Resolver(_maybe_throw)] = False

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        return Definitions()

    @classmethod
    def get_additional_scope(cls) -> Mapping[str, Any]:
        return {
            "error": _error,
        }


class MyNestedModel(BaseModel):
    a_string: str
    an_int: int

    model_config = ConfigDict(extra="forbid")


class MyNestedComponentSchema(BaseModel):
    nested: dict[str, MyNestedModel]

    model_config = ConfigDict(extra="forbid")


class MyNestedComponent(Component):
    @classmethod
    def get_schema(cls) -> type[MyNestedComponentSchema]:
        return MyNestedComponentSchema

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        return Definitions()



import shutil
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING

from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.decorators.asset_decorator import multi_asset
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster.components.component.component import Component
from dagster.components.core.context import ComponentLoadContext
from dagster.components.resolved.base import Resolvable
from dagster.components.resolved.core_models import ResolvedAssetSpec

if TYPE_CHECKING:
    from dagster._core.definitions.definitions_class import Definitions


@dataclass
class PipesSubprocessScript(Resolvable):
    path: str
    assets: Sequence[ResolvedAssetSpec]


@dataclass
class PipesSubprocessScriptCollectionComponent(Component, Resolvable):
    """Assets that wrap Python scripts executed with Dagster's PipesSubprocessClient."""

    scripts: Sequence[PipesSubprocessScript]

    @cached_property
    def specs_by_path(self) -> Mapping[str, Sequence[AssetSpec]]:
        return {script.path: script.assets for script in self.scripts}

    @staticmethod
    def introspect_from_path(path: Path) -> "PipesSubprocessScriptCollectionComponent":
        return PipesSubprocessScriptCollectionComponent(
            [
                PipesSubprocessScript(path=str(path), assets=[AssetSpec(path.stem)])
                for path in list(path.rglob("*.py"))
            ]
        )

    def build_defs(self, context: ComponentLoadContext) -> "Definitions":
        from dagster._core.definitions.definitions_class import Definitions

        return Definitions(
            assets=[
                self._create_asset_def(context.path / path, specs)
                for path, specs in self.specs_by_path.items()
            ],
        )

    def _create_asset_def(self, path: Path, specs: Sequence[AssetSpec]) -> AssetsDefinition:
        from dagster._core.pipes.subprocess import PipesSubprocessClient

        # TODO: allow name paraeterization
        @multi_asset(specs=specs, name=f"script_{path.stem}")
        def _asset(context: AssetExecutionContext):
            cmd = [shutil.which("python"), path]
            return PipesSubprocessClient().run(command=cmd, context=context).get_results()

        return _asset



from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from typing import Annotated, Any, Callable, Literal, Optional, Union

from dagster_shared.record import record
from typing_extensions import TypeAlias

import dagster._check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.asset_spec import AssetSpec, map_asset_specs
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.definitions_class import Definitions
from dagster._record import replace
from dagster.components.resolved.base import Resolvable, resolve_fields
from dagster.components.resolved.context import ResolutionContext
from dagster.components.resolved.model import Injectable, Injected, Model, Resolver


def _resolve_asset_key(context: ResolutionContext, key: str) -> AssetKey:
    resolved_val = context.resolve_value(key, as_type=AssetKey)
    return (
        AssetKey.from_user_string(resolved_val) if isinstance(resolved_val, str) else resolved_val
    )


PostProcessorFn: TypeAlias = Callable[[Definitions], Definitions]


class SingleRunBackfillPolicyModel(Resolvable, Model):
    type: Literal["single_run"] = "single_run"


class MultiRunBackfillPolicyModel(Resolvable, Model):
    type: Literal["multi_run"] = "multi_run"
    max_partitions_per_run: int = 1


def resolve_backfill_policy(
    context: ResolutionContext,
    backfill_policy,
) -> Optional[BackfillPolicy]:
    if backfill_policy is None:
        return None

    if backfill_policy.type == "single_run":
        return BackfillPolicy.single_run()
    elif backfill_policy.type == "multi_run":
        return BackfillPolicy.multi_run(
            max_partitions_per_run=backfill_policy.max_partitions_per_run
        )

    raise ValueError(f"Invalid backfill policy: {backfill_policy}")


@dataclass
class OpSpec(Resolvable):
    name: Optional[str] = None
    tags: Optional[dict[str, str]] = None
    backfill_policy: Annotated[
        Optional[BackfillPolicy],
        Resolver(
            resolve_backfill_policy,
            model_field_type=Union[SingleRunBackfillPolicyModel, MultiRunBackfillPolicyModel],
        ),
    ] = None


def _expect_injected(context, val):
    return check.opt_inst_param(val, "val", AutomationCondition)


ResolvedAssetKey = Annotated[
    AssetKey,
    Resolver(
        _resolve_asset_key,
        model_field_type=str,
    ),
]


@record
class SharedAssetKwargs(Resolvable):
    deps: Optional[Sequence[ResolvedAssetKey]] = None
    description: Optional[str] = None
    metadata: Injectable[Mapping[str, Any]] = {}
    group_name: Optional[str] = None
    skippable: Optional[bool] = None
    code_version: Optional[str] = None
    owners: Optional[Sequence[str]] = None
    tags: Injectable[Mapping[str, str]] = {}
    kinds: Sequence[str] = []
    automation_condition: Optional[Injected[AutomationCondition]] = None


@record
class AssetSpecKwargs(SharedAssetKwargs):
    key: ResolvedAssetKey


@record
class AssetAttributesKwargs(SharedAssetKwargs):
    key: Optional[ResolvedAssetKey] = None


def resolve_asset_spec(context: ResolutionContext, model):
    return AssetSpec(**resolve_fields(model, AssetSpecKwargs, context))


ResolvedAssetSpec: TypeAlias = Annotated[
    AssetSpec,
    Resolver(
        resolve_asset_spec,
        model_field_type=AssetSpecKwargs.model(),
    ),
]


def resolve_asset_attributes_to_mapping(
    context: ResolutionContext,
    model,
) -> Mapping[str, Any]:
    # only include fields that are explcitly set
    set_fields = model.model_dump(exclude_unset=True).keys()
    resolved_fields = resolve_fields(model, AssetAttributesKwargs, context)
    return {k: v for k, v in resolved_fields.items() if k in set_fields}


AssetAttributesModel = AssetAttributesKwargs.model()

ResolvedAssetAttributes: TypeAlias = Annotated[
    Mapping[str, Any],
    Resolver(
        resolve_asset_attributes_to_mapping,
        model_field_type=AssetAttributesKwargs.model(),
    ),
]


class AssetPostProcessorModel(Resolvable, Model):
    target: str = "*"
    operation: Literal["merge", "replace"] = "merge"
    attributes: ResolvedAssetAttributes


def apply_post_processor_to_spec(
    model: AssetPostProcessorModel, spec: AssetSpec, context: ResolutionContext
) -> AssetSpec:
    attributes = (
        context.with_scope(asset=spec).at_path("attributes").resolve_value(model.attributes)
    )
    if model.operation == "merge":
        mergeable_attributes = {"metadata", "tags"}
        merge_attributes = {k: v for k, v in attributes.items() if k in mergeable_attributes}
        replace_attributes = {k: v for k, v in attributes.items() if k not in mergeable_attributes}
        return spec.merge_attributes(**merge_attributes).replace_attributes(**replace_attributes)
    elif model.operation == "replace":
        return spec.replace_attributes(**attributes)
    else:
        check.failed(f"Unsupported operation: {model.operation}")


def apply_post_processor_to_defs(
    model: AssetPostProcessorModel, defs: Definitions, context: ResolutionContext
) -> Definitions:
    target_selection = AssetSelection.from_string(model.target, include_sources=True)
    target_keys = target_selection.resolve(defs.get_asset_graph())

    mappable = [d for d in defs.assets or [] if isinstance(d, (AssetsDefinition, AssetSpec))]
    mapped_assets = map_asset_specs(
        lambda spec: apply_post_processor_to_spec(model, spec, context)
        if spec.key in target_keys
        else spec,
        mappable,
    )

    assets = [
        *mapped_assets,
        *[d for d in defs.assets or [] if not isinstance(d, (AssetsDefinition, AssetSpec))],
    ]
    return replace(defs, assets=assets)


def resolve_schema_to_post_processor(
    context, model: AssetPostProcessorModel
) -> Callable[[Definitions], Definitions]:
    return lambda defs: apply_post_processor_to_defs(model, defs, context)


AssetPostProcessor: TypeAlias = Annotated[
    PostProcessorFn,
    Resolver(
        resolve_schema_to_post_processor,
        model_field_type=AssetPostProcessorModel,
    ),
]



import asyncio
from collections.abc import Sequence
from typing import Union

from dagster_shared.serdes import whitelist_for_serdes

from dagster._core.definitions.asset_key import T_EntityKey
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
    BuiltinAutomationCondition,
)
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._record import copy, record


@whitelist_for_serdes
@record
class SinceCondition(BuiltinAutomationCondition[T_EntityKey]):
    trigger_condition: AutomationCondition[T_EntityKey]
    reset_condition: AutomationCondition[T_EntityKey]

    @property
    def name(self) -> str:
        return "SINCE"

    @property
    def children(self) -> Sequence[AutomationCondition[T_EntityKey]]:
        return [self.trigger_condition, self.reset_condition]

    async def evaluate(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, context: AutomationContext[T_EntityKey]
    ) -> AutomationResult[T_EntityKey]:
        # must evaluate child condition over the entire subset to avoid missing state transitions
        child_candidate_subset = context.asset_graph_view.get_full_subset(key=context.key)

        # compute result for trigger and reset conditions
        trigger_result, reset_result = await asyncio.gather(
            *[
                context.for_child_condition(
                    self.trigger_condition,
                    child_indices=[0],
                    candidate_subset=child_candidate_subset,
                ).evaluate_async(),
                context.for_child_condition(
                    self.reset_condition,
                    child_indices=[1],
                    candidate_subset=child_candidate_subset,
                ).evaluate_async(),
            ]
        )

        # take the previous subset that this was true for
        true_subset = context.previous_true_subset or context.get_empty_subset()

        # add in any newly true trigger asset partitions
        true_subset = true_subset.compute_union(trigger_result.true_subset)
        # remove any newly true reset asset partitions
        true_subset = true_subset.compute_difference(reset_result.true_subset)

        return AutomationResult(
            context=context, true_subset=true_subset, child_results=[trigger_result, reset_result]
        )

    def replace(
        self, old: Union[AutomationCondition, str], new: AutomationCondition
    ) -> AutomationCondition:
        """Replaces all instances of ``old`` across any sub-conditions with ``new``.

        If ``old`` is a string, then conditions with a label matching
        that string will be replaced.

        Args:
            old (Union[AutomationCondition, str]): The condition to replace.
            new (AutomationCondition): The condition to replace with.
        """
        return (
            new
            if old in [self, self.get_label()]
            else copy(
                self,
                trigger_condition=self.trigger_condition.replace(old, new),
                reset_condition=self.reset_condition.replace(old, new),
            )
        )



from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Literal, Optional, TypeVar, Union

from pydantic import BaseModel
from typing_extensions import TypeAlias

from dagster import _check as check
from dagster._record import record

# Type variable for generic class handling
T = TypeVar("T")

# Constant for object attribute name
SCAFFOLDER_CLS_ATTRIBUTE = "__scaffolder_cls__"


def scaffold_with(
    scaffolder_cls: Union[type["Scaffolder"], "ScaffolderUnavailableReason"],
) -> Callable[[T], T]:
    """A decorator that declares what scaffolder is used to scaffold the artifact.

    Args:
        scaffolder_cls: A class that inherits from Scaffolder

    Returns:
        Decorator function that enhances the target class
    """
    from dagster.components.core.library_object import LIBRARY_OBJECT_ATTR

    def decorator(obj: T) -> T:
        # Store the scaffolder class as an attribute using the constant
        setattr(obj, SCAFFOLDER_CLS_ATTRIBUTE, scaffolder_cls)
        # All scaffoldable objects are library objects
        setattr(obj, LIBRARY_OBJECT_ATTR, True)
        return obj

    return decorator


def has_scaffolder(obj: object) -> bool:
    """Determines if an object has been decorated with `@scaffold_with`.

    Args:
        obj: The object to check

    Returns:
        True if the object has a Scaffolder attached, False otherwise
    """
    return hasattr(obj, SCAFFOLDER_CLS_ATTRIBUTE)


def get_scaffolder(
    obj: object,
) -> Union["Scaffolder", "ScaffolderUnavailableReason"]:
    """Retrieves the scaffolder class attached to the decorated object.

    Args:
        obj: The object to inspect

    Returns:
        The scaffolder class attached to the decorated object. Raises CheckError if the object is not decorated with @scaffold_with.
    """
    check.param_invariant(
        has_scaffolder(obj), "obj", "Object must be decorated with @scaffold_with"
    )
    attr = getattr(obj, SCAFFOLDER_CLS_ATTRIBUTE)
    return attr if isinstance(attr, ScaffolderUnavailableReason) else attr()


@dataclass
class ScaffolderUnavailableReason:
    message: str


ScaffoldFormatOptions: TypeAlias = Literal["yaml", "python"]


@record
class ScaffoldRequest:
    # fully qualified class name of the decorated object
    type_name: str
    # target path for the scaffold request. Typically used to construct absolute paths
    target_path: Path
    # yaml or python
    scaffold_format: ScaffoldFormatOptions


class Scaffolder:
    @classmethod
    def get_scaffold_params(cls) -> Optional[type[BaseModel]]:
        return None

    @abstractmethod
    def scaffold(self, request: ScaffoldRequest, params: Any) -> None: ...



from typing import TYPE_CHECKING, Union

from typing_extensions import TypeIs

from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)

if TYPE_CHECKING:
    from dagster._core.definitions.declarative_automation.operators import (
        AndAutomationCondition,
        ChecksAutomationCondition,
        DepsAutomationCondition,
        NotAutomationCondition,
        OrAutomationCondition,
    )


def has_allow_ignore(
    condition: AutomationCondition,
) -> TypeIs[
    Union[
        "AndAutomationCondition",
        "ChecksAutomationCondition",
        "DepsAutomationCondition",
        "NotAutomationCondition",
        "OrAutomationCondition",
    ]
]:
    from dagster._core.definitions.declarative_automation.operators import (
        AndAutomationCondition,
        ChecksAutomationCondition,
        DepsAutomationCondition,
        NotAutomationCondition,
        OrAutomationCondition,
    )

    return isinstance(
        condition,
        (
            AndAutomationCondition,
            ChecksAutomationCondition,
            DepsAutomationCondition,
            NotAutomationCondition,
            OrAutomationCondition,
        ),
    )



import inspect
from abc import abstractmethod
from collections.abc import Mapping, Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, TypeVar

from dagster_shared.serdes.objects import LibraryObjectKey
from dagster_shared.yaml_utils import parse_yaml_with_source_positions
from pydantic import BaseModel, ConfigDict, TypeAdapter

import dagster._check as check
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.module_loaders.load_defs_from_module import (
    load_definitions_from_module,
)
from dagster._core.definitions.module_loaders.utils import find_objects_in_module_of_types
from dagster._core.errors import DagsterInvalidDefinitionError
from dagster._utils import pushd
from dagster._utils.pydantic_yaml import (
    _parse_and_populate_model_with_annotated_errors,
    enrich_validation_errors_with_source_position,
)
from dagster.components.component.component import Component
from dagster.components.component.component_loader import is_component_loader
from dagster.components.core.context import ComponentLoadContext
from dagster.components.core.library_object import load_library_object
from dagster.components.resolved.base import Resolvable
from dagster.components.resolved.core_models import AssetPostProcessor

T = TypeVar("T", bound=BaseModel)


class ComponentFileModel(BaseModel):
    model_config = ConfigDict(extra="forbid")

    type: str
    attributes: Optional[Mapping[str, Any]] = None


#########
# MODULES
#########


@dataclass
class DefsModuleComponent(Component):
    path: Path

    @classmethod
    def from_context(cls, context: ComponentLoadContext) -> Optional["DefsModuleComponent"]:
        # this defines the priority of the module types
        module_types = (
            WrappedYamlComponent,
            WrappedPythonicComponent,
            DagsterDefsComponent,
            DefsFolderComponent,
        )
        module_filter = filter(None, (cls.from_context(context) for cls in module_types))
        return next(module_filter, None)

    @classmethod
    def load(cls, attributes: Any, context: ComponentLoadContext) -> "DefsModuleComponent":
        return check.not_none(cls.from_context(context))


@dataclass
class DefsFolderComponentYamlSchema(Resolvable):
    asset_post_processors: Optional[Sequence[AssetPostProcessor]] = None


@dataclass
class DefsFolderComponent(DefsModuleComponent):
    """A folder containing multiple submodules."""

    submodules: Sequence[DefsModuleComponent]
    asset_post_processors: Optional[Sequence[AssetPostProcessor]]

    @classmethod
    def get_schema(cls):
        return DefsFolderComponentYamlSchema.model()

    @classmethod
    def load(cls, attributes: Any, context: ComponentLoadContext) -> "DefsFolderComponent":
        # doing this funky thing because some of our attributes are resolved from the context,
        # so we split up resolving the yaml-defined attributes and the context-defined attributes,
        # meaning we manually invoke the resolution system here
        resolved_attributes = DefsFolderComponentYamlSchema.resolve_from_model(
            context.resolution_context.at_path("attributes"),
            attributes,
        )
        submodules = check.not_none(cls.submodules_from_context(context))
        return DefsFolderComponent(
            path=context.path,
            submodules=submodules,
            asset_post_processors=resolved_attributes.asset_post_processors,
        )

    @classmethod
    def submodules_from_context(
        cls, context: ComponentLoadContext
    ) -> Optional[Sequence[DefsModuleComponent]]:
        if not context.path.is_dir():
            return None
        submodules = (
            DefsModuleComponent.from_context(context.for_path(subpath))
            for subpath in context.path.iterdir()
        )
        return list(filter(None, submodules))

    @classmethod
    def from_context(cls, context: ComponentLoadContext) -> Optional["DefsFolderComponent"]:
        submodules = cls.submodules_from_context(context)
        if submodules is None:
            return None
        return DefsFolderComponent(
            path=context.path, submodules=submodules, asset_post_processors=None
        )

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        defs = Definitions.merge(
            *(
                submodule.build_defs(context.for_path(submodule.path))
                for submodule in self.submodules
            )
        )
        for post_processor in self.asset_post_processors or []:
            defs = post_processor(defs)
        return defs


@dataclass
class DagsterDefsComponent(DefsModuleComponent):
    """A module containing python dagster definitions."""

    @classmethod
    def from_context(cls, context: ComponentLoadContext) -> Optional["DagsterDefsComponent"]:
        if (context.path / "definitions.py").exists():
            return DagsterDefsComponent(path=context.path / "definitions.py")
        elif context.path.suffix == ".py":
            return DagsterDefsComponent(path=context.path)
        return None

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        module = context.load_defs_relative_python_module(self.path)
        definitions_objects = list(find_objects_in_module_of_types(module, Definitions))
        if len(definitions_objects) == 0:
            return load_definitions_from_module(module)
        elif len(definitions_objects) == 1:
            return next(iter(definitions_objects))
        else:
            raise DagsterInvalidDefinitionError(
                f"Found multiple Definitions objects in {self.path}. At most one Definitions object "
                "may be specified per module."
            )


@dataclass
class WrappedDefsModuleComponent(DefsModuleComponent):
    """A module containing a component definition."""

    wrapped: Component

    @staticmethod
    @abstractmethod
    def get_component_def_path(path: Path) -> Path: ...

    @classmethod
    @abstractmethod
    def get_component(cls, context: ComponentLoadContext) -> Component: ...

    @classmethod
    def from_context(cls, context: ComponentLoadContext) -> Optional["WrappedDefsModuleComponent"]:
        if cls.get_component_def_path(context.path).exists():
            return cls(path=context.path, wrapped=cls.get_component(context))
        return None

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        return self.wrapped.build_defs(context)


@dataclass
class WrappedPythonicComponent(WrappedDefsModuleComponent):
    @staticmethod
    def get_component_def_path(path: Path) -> Path:
        return path / "component.py"

    @classmethod
    def get_component(cls, context) -> Component:
        module = context.load_defs_relative_python_module(cls.get_component_def_path(context.path))
        component_loaders = list(inspect.getmembers(module, is_component_loader))
        if len(component_loaders) == 0:
            raise DagsterInvalidDefinitionError("No component loaders found in module")
        elif len(component_loaders) == 1:
            _, component_loader = component_loaders[0]
            return WrappedPythonicComponent(path=context.path, wrapped=component_loader(context))
        else:
            raise DagsterInvalidDefinitionError(
                f"Multiple component loaders found in module: {component_loaders}"
            )


@dataclass
class WrappedYamlComponent(WrappedDefsModuleComponent):
    @staticmethod
    def get_component_def_path(path: Path) -> Path:
        return path / "component.yaml"

    @classmethod
    def get_component(cls, context: ComponentLoadContext) -> Component:
        # parse the yaml file
        component_def_path = cls.get_component_def_path(context.path)
        source_tree = parse_yaml_with_source_positions(
            component_def_path.read_text(), str(component_def_path)
        )
        component_file_model = _parse_and_populate_model_with_annotated_errors(
            cls=ComponentFileModel, obj_parse_root=source_tree, obj_key_path_prefix=[]
        )

        # find the component type
        type_str = context.normalize_component_type_str(component_file_model.type)
        key = LibraryObjectKey.from_typename(type_str)
        obj = load_library_object(key)
        if not isinstance(obj, type) or not issubclass(obj, Component):
            raise DagsterInvalidDefinitionError(
                f"Component type {type_str} is of type {type(obj)}, but must be a subclass of dagster.Component"
            )

        component_schema = obj.get_schema()
        context = context.with_rendering_scope(
            obj.get_additional_scope()
        ).with_source_position_tree(source_tree.source_position_tree)

        # grab the attributes from the yaml file
        with pushd(str(context.path)):
            if component_schema is None:
                attributes = None
            elif source_tree:
                attributes_position_tree = source_tree.source_position_tree.children["attributes"]
                with enrich_validation_errors_with_source_position(
                    attributes_position_tree, ["attributes"]
                ):
                    attributes = TypeAdapter(component_schema).validate_python(
                        component_file_model.attributes
                    )
            else:
                attributes = TypeAdapter(component_schema).validate_python(
                    component_file_model.attributes
                )

        return obj.load(attributes, context)



import json
from typing import Literal, Optional, Union

import click
from dagster_shared.serdes.objects import LibraryObjectKey
from dagster_shared.serdes.serdes import serialize_value
from pydantic import ConfigDict, TypeAdapter, create_model

from dagster._cli.utils import assert_no_remaining_opts, get_possibly_temporary_instance_for_cli
from dagster._cli.workspace.cli_target import (
    PythonPointerOpts,
    get_repository_python_origin_from_cli_opts,
    python_pointer_options,
)
from dagster._core.definitions.asset_job import is_reserved_asset_job_name
from dagster._utils.hosted_user_process import recon_repository_from_origin
from dagster.components.component.component import Component
from dagster.components.core.defs import (
    DgAssetMetadata,
    DgDefinitionMetadata,
    DgJobMetadata,
    DgScheduleMetadata,
    DgSensorMetadata,
)
from dagster.components.core.library_object import (
    discover_entry_point_library_objects,
    discover_library_objects,
)
from dagster.components.core.snapshot import get_library_object_snap


@click.group(name="list")
def list_cli():
    """Commands for listing Dagster components and related entities."""


@list_cli.command(name="library")
@click.option("--entry-points/--no-entry-points", is_flag=True, default=True)
@click.argument("extra_modules", nargs=-1, type=str)
@click.pass_context
def list_library_command(
    ctx: click.Context, entry_points: bool, extra_modules: tuple[str, ...]
) -> None:
    """List registered library objects."""
    library_objects = _load_library_objects(entry_points, extra_modules)
    serialized_snaps = serialize_value(
        [get_library_object_snap(key, obj) for key, obj in library_objects.items()]
    )
    click.echo(serialized_snaps)


@list_cli.command(name="all-components-schema")
@click.option("--entry-points/--no-entry-points", is_flag=True, default=True)
@click.argument("extra_modules", nargs=-1, type=str)
def list_all_components_schema_command(entry_points: bool, extra_modules: tuple[str, ...]) -> None:
    """Builds a JSON schema which ORs the schema for a component
    file for all component types available in the current code location.
    """
    component_types = _load_component_types(entry_points, extra_modules)

    schemas = []
    for key in sorted(component_types.keys(), key=lambda k: k.to_typename()):
        component_type = component_types[key]
        # Create ComponentFileModel schema for each type
        schema_type = component_type.get_schema()
        key_string = key.to_typename()
        if schema_type:
            schemas.append(
                create_model(
                    key.name,
                    type=(Literal[key_string], key_string),
                    attributes=(schema_type, None),
                    __config__=ConfigDict(extra="forbid"),
                )
            )
    union_type = Union[tuple(schemas)]  # type: ignore
    click.echo(json.dumps(TypeAdapter(union_type).json_schema()))


@list_cli.command(name="definitions")
@python_pointer_options
@click.option(
    "--location", "-l", help="Name of the code location, can be used to scope environment variables"
)
@click.pass_context
def list_definitions_command(
    ctx: click.Context, location: Optional[str], **other_opts: object
) -> None:
    """List Dagster definitions."""
    python_pointer_opts = PythonPointerOpts.extract_from_cli_options(other_opts)
    assert_no_remaining_opts(other_opts)

    with get_possibly_temporary_instance_for_cli(
        "``dagster-components definitions list``"
    ) as instance:
        instance.inject_env_vars(location)

        repository_origin = get_repository_python_origin_from_cli_opts(python_pointer_opts)
        recon_repo = recon_repository_from_origin(repository_origin)
        repo_def = recon_repo.get_definition()

        all_defs: list[DgDefinitionMetadata] = []

        asset_graph = repo_def.asset_graph
        for key in sorted(
            list(asset_graph.get_all_asset_keys()), key=lambda key: key.to_user_string()
        ):
            node = asset_graph.get(key)
            all_defs.append(
                DgAssetMetadata(
                    type="asset",
                    key=key.to_user_string(),
                    deps=sorted([k.to_user_string() for k in node.parent_keys]),
                    group=node.group_name,
                    kinds=sorted(list(node.kinds)),
                    description=node.description,
                    automation_condition=node.automation_condition.get_label()
                    if node.automation_condition
                    else None,
                )
            )
        for job in repo_def.get_all_jobs():
            if not is_reserved_asset_job_name(job.name):
                all_defs.append(DgJobMetadata(type="job", name=job.name))
        for schedule in repo_def.schedule_defs:
            schedule_str = (
                schedule.cron_schedule
                if isinstance(schedule.cron_schedule, str)
                else ", ".join(schedule.cron_schedule)
            )
            all_defs.append(
                DgScheduleMetadata(
                    type="schedule",
                    name=schedule.name,
                    cron_schedule=schedule_str,
                )
            )
        for sensor in repo_def.sensor_defs:
            all_defs.append(DgSensorMetadata(type="sensor", name=sensor.name))

        click.echo(json.dumps(all_defs))


# ########################
# ##### HELPERS
# ########################


def _load_library_objects(
    entry_points: bool, extra_modules: tuple[str, ...]
) -> dict[LibraryObjectKey, object]:
    objects = {}
    if entry_points:
        objects.update(discover_entry_point_library_objects())
    if extra_modules:
        objects.update(discover_library_objects(extra_modules))
    return objects


def _load_component_types(
    entry_points: bool, extra_modules: tuple[str, ...]
) -> dict[LibraryObjectKey, type[Component]]:
    return {
        key: obj
        for key, obj in _load_library_objects(entry_points, extra_modules).items()
        if isinstance(obj, type) and issubclass(obj, Component)
    }



from abc import abstractmethod

from dagster.components.scaffold.scaffold import Scaffolder, ScaffoldRequest


class ShimScaffolder(Scaffolder):
    @abstractmethod
    def get_text(self, filename: str) -> str: ...
    def scaffold(self, request: ScaffoldRequest, params: None) -> None:
        if request.target_path.suffix != ".py":
            raise ValueError("Invalid target path suffix. Expected a path ending in `.py`.")
        # temporary hack as currently all scaffold requests target directories
        # that are auto-created
        request.target_path.rmdir()
        request.target_path.write_text(self.get_text(request.target_path.stem))



# re-export of dagster_shared.record to ease migration
# should be removed once call-sites migrated

from dagster_shared.record import (
    IHaveNew as IHaveNew,
    ImportFrom as ImportFrom,
    LegacyNamedTupleMixin as LegacyNamedTupleMixin,
    as_dict as as_dict,
    as_dict_for_new as as_dict_for_new,
    copy as copy,
    get_original_class as get_original_class,
    get_record_annotations as get_record_annotations,
    has_generated_new as has_generated_new,
    is_record as is_record,
    record as record,
    record_custom as record_custom,
    replace as replace,
)



from collections.abc import Sequence
from functools import update_wrapper
from typing import (  # noqa: UP035
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Optional,
    Union,
    cast,
    overload,
)

import dagster._check as check
from dagster._core.decorator_utils import get_function_params, validate_expected_params
from dagster._core.definitions.events import HookExecutionResult
from dagster._core.definitions.hook_definition import HookDefinition
from dagster._core.errors import DagsterInvalidDefinitionError

if TYPE_CHECKING:
    from dagster._core.events import DagsterEvent
    from dagster._core.execution.context.hook import HookContext


def _validate_hook_fn_params(fn, expected_positionals):
    params = get_function_params(fn)
    missing_positional = validate_expected_params(params, expected_positionals)
    if missing_positional:
        raise DagsterInvalidDefinitionError(
            f"'{fn.__name__}' decorated function does not have required positional "
            f"parameter '{missing_positional}'. Hook functions should only have keyword arguments "
            "that match input names and a first positional parameter named 'context' and "
            "a second positional parameter named 'event_list'."
        )


class _Hook:
    def __init__(
        self,
        name: Optional[str] = None,
        required_resource_keys: Optional[AbstractSet[str]] = None,
        decorated_fn: Optional[Callable[..., Any]] = None,
    ):
        self.name = check.opt_str_param(name, "name")
        self.required_resource_keys = check.opt_set_param(
            required_resource_keys, "required_resource_keys"
        )
        self.decorated_fn = check.opt_callable_param(decorated_fn, "decorated_fn")

    def __call__(self, fn) -> HookDefinition:
        check.callable_param(fn, "fn")

        if not self.name:
            self.name = fn.__name__

        expected_positionals = ["context", "event_list"]

        _validate_hook_fn_params(fn, expected_positionals)

        hook_def = HookDefinition(
            name=self.name or "",
            hook_fn=fn,
            required_resource_keys=self.required_resource_keys,
            decorated_fn=self.decorated_fn or fn,
        )
        update_wrapper(cast(Callable[..., Any], hook_def), fn)
        return hook_def


@overload
def event_list_hook(
    hook_fn: Callable[..., Any],
) -> HookDefinition:
    pass


@overload
def event_list_hook(
    *,
    name: Optional[str] = ...,
    required_resource_keys: Optional[AbstractSet[str]] = ...,
    decorated_fn: Optional[Callable[..., Any]] = ...,
) -> _Hook:
    pass


def event_list_hook(
    hook_fn: Optional[Callable] = None,
    *,
    name: Optional[str] = None,
    required_resource_keys: Optional[AbstractSet[str]] = None,
    decorated_fn: Optional[Callable[..., Any]] = None,
) -> Union[HookDefinition, _Hook]:
    """Create a generic hook with the specified parameters from the decorated function.

    This decorator is currently used internally by Dagster machinery to support success_hook and
    failure_hook.

    The user-defined hook function requires two parameters:
    - A `context` object is passed as the first parameter. The context is an instance of
        :py:class:`context <HookContext>`, and provides access to system
        information, such as loggers (context.log), resources (context.resources), the op
        (context.op) and its execution step (context.step) which triggers this hook.
    - An `event_list` object is passed as the second parameter. It provides the full event list of the
        associated execution step.

    Args:
        name (Optional[str]): The name of this hook.
        required_resource_keys (Optional[AbstractSet[str]]): Keys for the resources required by the
            hook.

    Examples:
        .. code-block:: python

            @event_list_hook(required_resource_keys={'slack'})
            def slack_on_materializations(context, event_list):
                for event in event_list:
                    if event.event_type == DagsterEventType.ASSET_MATERIALIZATION:
                        message = f'{context.op_name} has materialized an asset {event.asset_key}.'
                        # send a slack message every time a materialization event occurs
                        context.resources.slack.send_message(message)


    """
    # This case is for when decorator is used bare, without arguments.
    # e.g. @event_list_hook versus @event_list_hook()
    if hook_fn is not None:
        check.invariant(required_resource_keys is None)
        return _Hook()(hook_fn)

    return _Hook(
        name=name, required_resource_keys=required_resource_keys, decorated_fn=decorated_fn
    )


SuccessOrFailureHookFn = Callable[["HookContext"], Any]


@overload
def success_hook(hook_fn: SuccessOrFailureHookFn) -> HookDefinition: ...


@overload
def success_hook(
    *,
    name: Optional[str] = ...,
    required_resource_keys: Optional[AbstractSet[str]] = ...,
) -> Callable[[SuccessOrFailureHookFn], HookDefinition]: ...


def success_hook(
    hook_fn: Optional[SuccessOrFailureHookFn] = None,
    *,
    name: Optional[str] = None,
    required_resource_keys: Optional[AbstractSet[str]] = None,
) -> Union[HookDefinition, Callable[[SuccessOrFailureHookFn], HookDefinition]]:
    """Create a hook on step success events with the specified parameters from the decorated function.

    Args:
        name (Optional[str]): The name of this hook.
        required_resource_keys (Optional[AbstractSet[str]]): Keys for the resources required by the
            hook.

    Examples:
        .. code-block:: python

            @success_hook(required_resource_keys={'slack'})
            def slack_message_on_success(context):
                message = 'op {} succeeded'.format(context.op.name)
                context.resources.slack.send_message(message)

            @success_hook
            def do_something_on_success(context):
                do_something()


    """

    def wrapper(fn: SuccessOrFailureHookFn) -> HookDefinition:
        check.callable_param(fn, "fn")

        expected_positionals = ["context"]
        _validate_hook_fn_params(fn, expected_positionals)

        if name is None or callable(name):
            _name = fn.__name__
        else:
            _name = name

        @event_list_hook(name=_name, required_resource_keys=required_resource_keys, decorated_fn=fn)
        def _success_hook(
            context: "HookContext", event_list: Sequence["DagsterEvent"]
        ) -> HookExecutionResult:
            for event in event_list:
                if event.is_step_success:
                    fn(context)
                    return HookExecutionResult(hook_name=_name, is_skipped=False)

            # hook is skipped when fn didn't run
            return HookExecutionResult(hook_name=_name, is_skipped=True)

        return _success_hook

    # This case is for when decorator is used bare, without arguments, i.e. @success_hook
    if hook_fn is not None:
        check.invariant(required_resource_keys is None)
        return wrapper(hook_fn)

    return wrapper


@overload
def failure_hook(name: SuccessOrFailureHookFn) -> HookDefinition: ...


@overload
def failure_hook(
    name: Optional[str] = ...,
    required_resource_keys: Optional[AbstractSet[str]] = ...,
) -> Callable[[SuccessOrFailureHookFn], HookDefinition]: ...


def failure_hook(
    name: Optional[Union[SuccessOrFailureHookFn, str]] = None,
    required_resource_keys: Optional[AbstractSet[str]] = None,
) -> Union[HookDefinition, Callable[[SuccessOrFailureHookFn], HookDefinition]]:
    """Create a hook on step failure events with the specified parameters from the decorated function.

    Args:
        name (Optional[str]): The name of this hook.
        required_resource_keys (Optional[AbstractSet[str]]): Keys for the resources required by the
            hook.

    Examples:
        .. code-block:: python

            @failure_hook(required_resource_keys={'slack'})
            def slack_message_on_failure(context):
                message = 'op {} failed'.format(context.op.name)
                context.resources.slack.send_message(message)

            @failure_hook
            def do_something_on_failure(context):
                do_something()


    """

    def wrapper(fn: Callable[["HookContext"], Any]) -> HookDefinition:
        check.callable_param(fn, "fn")

        expected_positionals = ["context"]
        _validate_hook_fn_params(fn, expected_positionals)

        if name is None or callable(name):
            _name = fn.__name__
        else:
            _name = name

        @event_list_hook(name=_name, required_resource_keys=required_resource_keys, decorated_fn=fn)
        def _failure_hook(
            context: "HookContext", event_list: Sequence["DagsterEvent"]
        ) -> HookExecutionResult:
            for event in event_list:
                if event.is_step_failure:
                    fn(context)
                    return HookExecutionResult(hook_name=_name, is_skipped=False)

            # hook is skipped when fn didn't run
            return HookExecutionResult(hook_name=_name, is_skipped=True)

        return _failure_hook

    # This case is for when decorator is used bare, without arguments, i.e. @failure_hook
    if callable(name):
        check.invariant(required_resource_keys is None)
        return wrapper(name)

    return wrapper



from typing import Literal, Optional, TypedDict, Union

from typing_extensions import TypeAlias

DgDefinitionMetadata: TypeAlias = Union[
    "DgAssetMetadata", "DgSensorMetadata", "DgScheduleMetadata", "DgJobMetadata"
]


class DgAssetMetadata(TypedDict):
    type: Literal["asset"]
    key: str
    deps: list[str]
    kinds: list[str]
    group: Optional[str]
    description: Optional[str]
    automation_condition: Optional[str]


class DgSensorMetadata(TypedDict):
    type: Literal["sensor"]
    name: str


class DgScheduleMetadata(TypedDict):
    type: Literal["schedule"]
    name: str
    cron_schedule: str


class DgJobMetadata(TypedDict):
    type: Literal["job"]
    name: str



from dagster.components.core.defs_module import DefsFolderComponent as DefsFolderComponent
from dagster.components.lib.definitions_component import (
    DefinitionsComponent as DefinitionsComponent,
)
from dagster.components.lib.pipes_subprocess_script_collection import (
    PipesSubprocessScriptCollectionComponent as PipesSubprocessScriptCollectionComponent,
)

# These just modify the exsiting Dagster decorators
from dagster.components.lib.shim_components.asset import asset as asset
from dagster.components.lib.shim_components.schedule import schedule as schedule
from dagster.components.lib.shim_components.sensor import sensor as sensor



from collections.abc import Mapping
from typing import Any, TypeVar

from dagster_shared.merger import deep_merge_dicts as deep_merge_dicts

import dagster._check as check

K = TypeVar("K")
V = TypeVar("V")
K2 = TypeVar("K2")
V2 = TypeVar("V2")


def merge_dicts(*args: Mapping[Any, Any]) -> dict[Any, Any]:
    """Returns a dictionary with with all the keys in all of the input dictionaries.

    If multiple input dictionaries have different values for the same key, the returned dictionary
    contains the value from the dictionary that comes latest in the list.
    """
    check.is_tuple(args, of_type=dict)
    if len(args) < 2:
        check.failed(f"Expected 2 or more args to merge_dicts, found {len(args)}")

    result: dict[object, object] = {}
    for arg in args:
        result.update(arg)
    return result


def reverse_dict(d: Mapping[V, K]) -> dict[K, V]:
    """Returns a new dictionary with the keys and values of the input dictionary swapped.

    If the input dictionary has duplicate values, the returned dictionary will have the value from
    the last key that maps to it.
    """
    check.dict_param(d, "d")
    return {v: k for k, v in d.items()}



import inspect
from abc import ABC, abstractmethod
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel
from typing_extensions import Self

from dagster._core.definitions.definitions_class import Definitions
from dagster.components.component.component_scaffolder import DefaultComponentScaffolder
from dagster.components.resolved.base import Resolvable
from dagster.components.scaffold.scaffold import scaffold_with

if TYPE_CHECKING:
    from dagster.components.core.context import ComponentLoadContext


@scaffold_with(DefaultComponentScaffolder)
class Component(ABC):
    @classmethod
    def __dg_library_object__(cls) -> None: ...

    @classmethod
    def get_schema(cls) -> Optional[type[BaseModel]]:
        if issubclass(cls, Resolvable):
            return cls.model()

        return None

    @classmethod
    def get_additional_scope(cls) -> Mapping[str, Any]:
        return {}

    @abstractmethod
    def build_defs(self, context: "ComponentLoadContext") -> Definitions: ...

    @classmethod
    def load(cls, attributes: Optional[BaseModel], context: "ComponentLoadContext") -> Self:
        if issubclass(cls, Resolvable):
            return (
                cls.resolve_from_model(
                    context.resolution_context.at_path("attributes"),
                    attributes,
                )
                if attributes
                else cls()
            )
        else:
            # If the Component does not implement anything from Resolved, try to instantiate it without
            # argument.
            return cls()

    @classmethod
    def get_description(cls) -> Optional[str]:
        return inspect.getdoc(cls)



import sys
import traceback
from dataclasses import dataclass
from typing import TYPE_CHECKING, Annotated, Any, Callable, Optional, TypeVar, Union

from pydantic import BaseModel, ConfigDict

from dagster import _check as check

try:
    # this type only exists in python 3.10+
    from types import UnionType  # type: ignore
except ImportError:
    UnionType = Union

if TYPE_CHECKING:
    from dagster.components.resolved.context import ResolutionContext


class Model(BaseModel):
    """pydantic BaseModel configured to disallow extra fields in order to help catch errors earlier."""

    model_config = ConfigDict(extra="forbid")


def _recurse(context: "ResolutionContext", field_value):
    return context.resolve_value(field_value)


@dataclass
class ParentFn:
    callable: Callable[["ResolutionContext", Any], Any]


@dataclass
class AttrWithContextFn:
    callable: Callable[["ResolutionContext", Any], Any]


class Resolver:
    """Contains information on how to resolve a field from a model."""

    def __init__(
        self,
        fn: Union[ParentFn, AttrWithContextFn, Callable[["ResolutionContext", Any], Any]],
        model_field_name: Optional[str] = None,
        model_field_type: Optional[type] = None,
        can_inject: bool = False,
    ):
        """Resolve this field by invoking the function which will receive the corresponding field value
        from the model.
        """
        if not isinstance(fn, (ParentFn, AttrWithContextFn)):
            if not callable(fn):
                check.param_invariant(
                    callable(fn),
                    "fn",
                    f"must be callable if not ParentFn or AttrWithContextFn. Got {fn}",
                )
            self.fn = AttrWithContextFn(fn)
        else:
            self.fn = fn

        self.model_field_name = model_field_name
        self.model_field_type = model_field_type
        self.can_inject = can_inject

        super().__init__()

    @staticmethod
    def from_model(fn: Callable[["ResolutionContext", Any], Any], **kwargs):
        """Resolve this field by invoking the function which will receive the entire parent model."""
        return Resolver(ParentFn(fn), **kwargs)

    @staticmethod
    def default(**kwargs):
        """Default recursive resolution."""
        return Resolver(_recurse, **kwargs)

    def execute(
        self,
        context: "ResolutionContext",
        model: BaseModel,
        field_name: str,
    ) -> Any:
        from dagster.components.resolved.context import ResolutionException

        try:
            if isinstance(self.fn, ParentFn):
                return self.fn.callable(context, model)
            elif isinstance(self.fn, AttrWithContextFn):
                field_name = self.model_field_name or field_name
                attr = getattr(model, field_name)
                context = context.at_path(field_name)
                return self.fn.callable(context, attr)
        except ResolutionException:
            raise  # already processed
        except Exception:
            raise context.build_resolve_fn_exc(
                traceback.format_exception(*sys.exc_info()),
                field_name=field_name,
                model=model,
            ) from None

        raise ValueError(f"Unsupported Resolver type: {self.fn}")


T = TypeVar("T")

Injectable = Annotated[T, Resolver.default(can_inject=True)]
Injected = Annotated[T, Resolver.default(model_field_type=str)]
